(*
Copyright (C) 2000 Chris Teague
Copyright (C) 2000 Chris Millward (Chris.R.Millward@btinternet.com)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
*)

unit ufrmCMNES;

interface

uses
	Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
	Menus, DirectDraw, StdCtrls, ExtCtrls, ComCtrls, Registry, MMSystem,
	DirectInput, DxCommon;

type
	TGamePad = Object
		private
			{}
		public
			PadInput:				Integer;
			ResetValue:			Integer;
			Strobe:					Byte;
			ReadNum:        Byte;
			PadNumber:			Integer;
			Constructor Create;
			Procedure SetPad(Port: Integer);
			Function ReadByte: Byte;
			Procedure WriteByte(Address: Word;Value: Byte);
		end;

type
	PIDirectInputEffect = ^IDirectInputEffect;

type
	TInput = Object
		private
			InputType:			Byte;
			pDI:						IDIRECTINPUT7;
			pDIKeyboard:    IDIRECTINPUTDEVICE;
			Keystate:				Array [0..255] of Byte;
			pDIJoystick:  	IDirectInputDevice2;
			JoyList:				TStringList;
			JoyState:				TDIJOYSTATE2;
			Effect1:        TDIEFFECT;
			pConstantForce:	IDIRECTINPUTEFFECT;
			dwAxes:					Array [0..1] of DWORD;
			direction:			Array [0..1] of LongInt;
			diCF:						TDICONSTANTFORCE;
			diEnv:					TDIENVELOPE;
		public
			constructor Create;
			destructor Destroy;
			function ReadPad(Pad: Byte;Read: Byte): Byte;
			function CreateEffects: Boolean;
		end;

type
	TMemPointer = ^TByteArray;

type
	TGenericMemory = Object
		private
			PMemory:							TMemPointer;	//pointer to start of memory
		public
			MemorySize:	LongInt;
			function ReadByte(Address: Word): Byte;
			procedure WriteByte(Address: Word;Value: Byte);
			Constructor Create(Size: LongInt);
			Destructor Destroy;
		end;

type
	TROMHeader = Object
		private
			{}
		public
			PrgSize:							Integer;
			ChrSize:							Integer;
			Mapper:								Byte;	//number of the mapper
			BatteryBackup:				Byte;	//has battery backed up SRAM 1=yes, 0=no
			Trainer:							Byte; //ROM file has a trainer 1=yes, 0=no
			FourScreenMirror:			Byte; //4 screen mirroring 1=yes, 0=no
			Mirroring:						Byte; //1 = Vertical, 0 = Horizontal
		end;

Type
	TMapperA = Object
		private
			{}
		public
			{mapper 1}
			M1Array: Array [0..4] of Byte;
			M1Regs:	Array [0..3] of Byte;
			M1HighBank: Integer;
			M1LowBank:  Integer;
			DecodedLow:	Cardinal;
			DecodedHigh:Cardinal;
			M1Counter:	Byte;
			Constructor Create(MAROM:	TMemPointer;MAVROM: TMemPointer;MASRAM: TMemPointer;ImageSize:	LongInt);
			Destructor Destroy;
			function ReadByte(Address: Word): Byte;
			procedure WriteByte(Address: Word;Value: Byte);
		end;

type
	TTile = Object
		private
			Colour: Array [0..7] of Array [0..7] of Byte;
		public
			Constructor Create;
			Destructor Destroy;
			procedure SetColour(Row: Byte;Col: Byte;Value: Byte);
			function GetColor(Row: Byte;Col: Byte): Byte;
		end;

type
	TPPU = Object
		private
    	{}
		public
    	BackgroundColor:			Byte;
    	UpdateBackground: 		Byte;
			PPURegs:							array [0..8] of Byte;
			ImagePalette: 				array [0..16] of Byte;
			SpritePalette: 				array [0..16] of Byte;
			PPUMirroring:					Byte;
			BackScrollHorizontal:	Byte;
			BackScrollVertical:		Byte;
			VRAMAddrReg:					Word;
			VRAMAddrRegBuff:			Word;
			VRAMDataBuff:					Byte;
			bSkipFrame:						Boolean;
			TablesToUpdate:				Byte;
			bTable0Done:					Boolean;
			bTable1Done:					Boolean;
			bTable2Done:					Boolean;
			bTable3Done:					Boolean;
			PtnTableUpdated: 			Byte;
			SpriteRam:						array [0..256] of Byte;
			PVRAM:								TMemPointer;
			MyTableLow: 					array [0..256] of TTile;
			MyTableHigh: 					array [0..256] of TTile;
			Procedure FillPatternTableCache;
			Procedure VRAMWriteByte(Address: Word;Value: Byte);
			Function VRAMReadByte(Address: Word): Byte;
			Procedure DrawFrame();
			Procedure DrawBackground();
			procedure DrawTile(QuadrantNum: Word; TableNum, TileRow: Byte; TileColumn: Word; TileIndex: Byte);
			Procedure DrawSprite(Y: Byte;X: Byte;TileIndex: Byte;Attributes: Byte;PtnTableAddr: Word);
			Procedure Draw8x16Sprite(Y: Byte; X: Byte;TileIndex: Byte;Attributes: Byte);
			Function GetAttribColor(TableNum: Byte;Row: Byte;Col: Byte): Byte;
			Constructor Create(PASRAM: TGenericMemory;Mirroring: Byte;PASMapper: TMapperA);
			Destructor Destroy;
			Procedure SetValues(PASRAM: TGenericMemory;PASVRAM: TGenericMemory;Mirroring: Byte; PASMapper: TMapperA);
			Procedure StartVBlank();
			Procedure	EndVBlank();
			Function ReadByte(Address: Word): Byte;
			Procedure WriteByte(Address: Word;Value: Byte);
			Procedure DisplayPatternTable();
			Procedure DisplayPalette();
		end;

type
	TMemBlock = Array[0..65535] of Byte;

type
	TpAPU = Object
		private
			{}
		public
			constructor Create();
			destructor Destroy();
			function Initialise(): Boolean;
			function CH1WriteByte(Value: Byte): Boolean;
		end;

type
	TMemBus = Object
		private
			{}
		public
			Constructor Create(BIRAM: TGenericMemory;BMapper: TMapperA;BPPU: TPPU);
			Destructor Destroy;
			function ReadByte(Address: Word): Byte;
			procedure WriteByte(Address: Word;Value: Byte);
			function ReadAddress(Address: Word): Word;
		end;

type
	TCPU = Object
		private
			RegAcc:				Byte;					//	Register acc
			RegX:					Byte;					//	Register x
			RegY:					Byte;					//	Register y
			PC:						Word;					//	Program counter
			IR:						Byte;					//	Instruction register
			SR:						Byte;					//	Status register
			SP:						Word;					//	Stack pointer
			function BitSet(BitNumber: ShortInt) : Byte;
			Procedure SetCMPFlags(Value1: Byte;Value2: Byte);
			Procedure PushByte(Value: Byte);
			Procedure PushWord(Value: Word);
			function PopByte(): Byte;
			//function PopWord():	Word;
		public
			Constructor Create(MemoryBus: TMemBus);
			Destructor Destroy;
			Procedure Reset;									//	Reset the 6502
			function ExecuteInstruction(var Cycles: Int64): Byte;			//	Execute next 6502 instruction
			Procedure SetBus(MemoryBus: TMemBus);
			Procedure VBlankInterrupt(Cycles: Int64);
			Procedure SRSetBit(Position: Integer);
			Procedure SRClearBit(Position: Integer);
			Procedure SRSetNZ(Value: Byte);
			Procedure SRSetN(Value: Byte);
			Procedure SRSetZ(Value: Byte);
//------------------------------------------------------------
// 6502 address modes.
//------------------------------------------------------------
			function AbsoluteOperandAddress(): Word;
			function ImmediateOperandAddress(): Word;
			function IndirectOperandAddress(): Word;
			function AbsoluteXOperandAddress(): Word;
			function AbsoluteYOperandAddress(): Word;
			function ZeroPageOperandAddress(): Word;
			function ZeroPageXOperandAddress(): Word;
			function ZeroPageYOperandAddress(): Word;
			function IndirectYOperandAddress(): Word;
			function IndirectXOperandAddress(): Word;
			function RelativeOperandAddress(): Word;
		end;

type TGraphics = Object
		pDD:						IDIRECTDRAW;
		g_pDD:					IDIRECTDRAW4;
		g_pDDSPrimary:	IDIRECTDRAWSURFACE4;
		g_pDDSBack:			IDIRECTDRAWSURFACE4;
		g_pDDSNESBack:	IDIRECTDRAWSURFACE4;
		g_pClipper:			IDIRECTDRAWCLIPPER;
		WinHandle:			HWND;
		FullScreen:			Boolean;
		lpddpal:				IDIRECTDRAWPALETTE;
		g_rcViewport:		TRect;
		g_rcScreen:			TRect;
		ddsdsetpixel:		TDDSURFACEDESC2;
		ddsdsetpixelb:  TDDSURFACEDESC2;
		VidBuf:					^TByteArray;
		BackVidBuf:			^TByteArray;
	private
		{}
	public
		constructor Create;
		destructor Destroy;
		function Initialise(Handle: HWND): Boolean;
		function CreateSurfaces(bFullScreen: Boolean): Boolean;
		procedure CheckSurfaces;
		function ClearSurface(Mode: Integer): Boolean;
		procedure SetPixel(Surface: IDirectDrawSurface4;x, y: Integer; Color: Byte);
		procedure SetPixelBehindBackground(x, y: Integer; Color, BackgroundColor: Byte);
		procedure BackSetPixel(Surface: IDirectDrawSurface4;x, y: Integer; Color: Byte);
		procedure DrawFrame;
		function BlitNESBackground(HorizontalScroll,VerticalScroll: Byte): Boolean;
		procedure InitPalette;
		procedure Flip;
		procedure LockBackground(Surface: IDirectDrawSurface4);
		procedure LockNESBackground(Surface: IDirectDrawSurface4);
		procedure UnlockSurface(Surface: IDirectDrawSurface4);
	end;


type
	TfrmCMNES = class(TForm)
		mmnuCMNES: TMainMenu;
		mnuFile: TMenuItem;
		mnuOpenROM: TMenuItem;
		mnuCloseROM: TMenuItem;
		N1: TMenuItem;
		mnuExit: TMenuItem;
		mnuOptions: TMenuItem;
		mnuExecute: TMenuItem;
		mnuDrawBackground: TMenuItem;
		mnuDrawSprites: TMenuItem;
		N2: TMenuItem;
		mnuSound: TMenuItem;
		mnuGo: TMenuItem;
		mnuHalt: TMenuItem;
		mnuHelp: TMenuItem;
		mnuAbout: TMenuItem;
		N3: TMenuItem;
		mnuDebug: TMenuItem;
		mnuMemory: TMenuItem;
		dlgOpenROM: TOpenDialog;
		mnuROMInfo: TMenuItem;
		mnuClearScreen: TMenuItem;
		N4: TMenuItem;
		mnuCPUReset: TMenuItem;
		mnuWindowed: TMenuItem;
		mnuDisplayPatternTable: TMenuItem;
		mnuDisplayPalette: TMenuItem;
		mnuEventLog: TMenuItem;
		imgTitle: TImage;
		N7: TMenuItem;
		mnuCaptureScreen: TMenuItem;
		tmrFramesPerSecond: TTimer;
		mnuFramesPerSecond: TMenuItem;
		N8: TMenuItem;
		N5: TMenuItem;
		mnuSettings: TMenuItem;
    tmrDIPoll: TTimer;
		procedure mnuExitClick(Sender: TObject);
		procedure mnuMemoryClick(Sender: TObject);
		procedure mnuOpenROMClick(Sender: TObject);
		function ReadROMHeader: Boolean;
		function ReadROM(PrgSize: Integer; ROM: TGenericMemory;
			ChrSize: Integer; VROM: TGenericMemory; Trainer: Byte): Boolean;
		procedure CreateNES;
		procedure DestroyNES;
		procedure mnuROMInfoClick(Sender: TObject);
		procedure mnuCloseROMClick(Sender: TObject);
		procedure mnuGoClick(Sender: TObject);
		procedure mnuHaltClick(Sender: TObject);
		procedure FormCreate(Sender: TObject);
		procedure mnuDrawBackgroundClick(Sender: TObject);
		procedure mnuDrawSpritesClick(Sender: TObject);
		procedure mnuClearScreenClick(Sender: TObject);
		procedure mnuCPUResetClick(Sender: TObject);
		procedure mnuWindowedClick(Sender: TObject);
		procedure mnuDisplayPatternTableClick(Sender: TObject);
		procedure mnuDisplayPaletteClick(Sender: TObject);
		procedure mnuEventLogClick(Sender: TObject);
		procedure mnuDebugClick(Sender: TObject);
		procedure UpdateDebug;
		procedure ExecuteStep;
		procedure ClickAddress(Sender: TLabel);
		procedure mnuCaptureScreenClick(Sender: TObject);
		procedure mnuAboutClick(Sender: TObject);
		procedure tmrFramesPerSecondTimer(Sender: TObject);
		procedure mnuFramesPerSecondClick(Sender: TObject);
		procedure mnuSettingsClick(Sender: TObject);
		procedure GetRegistrySettings;
		procedure SetRegistrySettings;
		procedure FormClose(Sender: TObject; var Action: TCloseAction);
		procedure SetFrameSkip;
    procedure FormShow(Sender: TObject);
    procedure tmrDIPollTimer(Sender: TObject);
	private
		hWindow: HWND;
		{ Private declarations }
	public
		procedure DXError(HR: HRESULT);
	end;

const
	STRREGKEY								= '\Software\CMNES\';
	NMI_VECTOR: Word				= $FFFA;
	RESET_VECTOR:	Word			= $FFFC;
	IRQ_BRK_VECTOR:	Word		= $FFFE;

//bit numbers for the flags in the Status Register
	SR_CARRY_FLAG    = 0;  // carry
	SR_ZERO_FLAG     = 1;  // zero
	SR_INTR_FLAG     = 2;  // interrupts disabled
	SR_DECIMAL_FLAG  = 3;  // decimal mode
	SR_BREAK_FLAG    = 4;  // break
	SR_RSV_FLAG      = 5;  // reserved, always 1
	SR_OVERFLOW_FLAG = 6;  // overflow occurred
	SR_NEG_FLAG      = 7;  // negative

	HBLANK 	= 115;
	SCANLINES = 240;
	VPERIOD = 2875;

var
	frmCMNES: 							TfrmCMNES;
  {NES Components}
	NESScreen:							TGraphics;
	PPU:										TPPU;
	ROM:										TGenericMemory;
	VROM:										TGenericMemory;
	ROMHeader:							TRomHeader;
	SRAM:										TGenericMemory;
	IRAM:										TGenericMemory;
	Mapper0:								TMapperA;
	MEMBUS:									TMemBus;
	CPU:										TCPU;
	Input:									TInput;
	GamePad1:								TGamePad;
	GamePad2:								TGamePad;
	VRAM:										TGenericMemory;
	pAPU:										TpAPU;
	{Variables}
	bExit:									Boolean;
	ScanLine:								Integer;
	PatternTableDisplayed:	Boolean;
	bBackgroundEnabled:			Boolean;
	bSpritesEnabled:				Boolean;
	mmoEvents:							TMemo;
	CycleCount:							Int64;
	FrameCount:							Integer;
	TotalCycles:						Int64;
	strROMFileName:					String;
	iStartTime:							Cardinal;
	iEndTime:								Cardinal;
	strROMPath:							String;
	strScreenCapturePath:		String;
	strSoundCapturePath:		String;
	strSavePath:						String;
	strDisplayMode:					String;
	iSkip:									Integer; {mode}
	iFrameSkip:							Integer; {counter}
	Screen_Width:						Integer;
	Screen_Height:					Integer;
	strFrameCount:					String;
	strCPUSpeed:						String;
	{direct input}
	FDataFormat: TDIDataFormat;
	FDataFormatObjects: array[0..255] of TDIObjectDataFormat;
	FDataFormatGUIDs: array[0..255] of TGUID;
  FAxisList: array[0..255] of Integer;
	NumAxes: Integer;

implementation

uses ufrmMemory, ufrmEventLog, ufrmROMInfo, ufrmDebug, ufrmAbout,
	ufrmSettings;

{$R *.DFM}
//***************************//
//	TGenericRAM              //
//***************************//

function EnumInputDevs(var lpddi: TDIDeviceInstance; pvRef: Pointer): Integer; stdcall;
var
  TempGuid: PGuid;
begin
	{we must save this guid in a seperate list for use later when we
   create the device}
  New(TempGuid);
  TempGuid^ := lpddi.guidInstance;

	{display device information}
	TStringList(pvRef).AddObject(lpddi.tszInstanceName+' - '+lpddi.tszProductName, TObject(TempGuid));

  {continue enumeration}
	Result := Integer(True);
end;

procedure SetDataFormat;

  function DIEnumDeviceObjectsProc(const peff: TDIDeviceObjectInstance;
    pvRef: Pointer): HRESULT; stdcall;
  begin
    {if the offset is not beyond the indicated data size... (useful if
     we were to use the TDIJoyState data structure instead) }
    if peff.dwOfs<FDataFormat.dwDataSize then
    begin
      {save the type GUID for this object}
      FDataFormatGUIDs[FDataFormat.dwNumObjs] := peff.guidType;

      {add the object information to our custom made
       data format}
      with FDataFormatObjects[FDataFormat.dwNumObjs] do
      begin
        pguid := @FDataFormatGUIDs[FDataFormat.dwNumObjs];
        dwOfs := peff.dwOfs;
        dwType := peff.dwType;
        dwFlags := 0;
      end;

      {if this object is an axis, store it in our axis list}
      if (peff.dwType and DIDFT_AXIS) > 0 then
      begin
        FAxisList[NumAxes] := peff.dwOfs;
        Inc(NumAxes);
      end;

      {increment the number of objects found}
      Inc(FDataFormat.dwNumObjs);
    end;

    {continue enumeration}
    Result := DIENUM_CONTINUE;
  end;

begin
	{enumerate all objects on this input device. this will build our
	 custom data structure programmatically with the correct information
	 for each object}
	Input.pDIJoystick.EnumObjects(@DIEnumDeviceObjectsProc, nil, DIDFT_ALL);

	{set the data format for the input device}
	Input.pDIJoystick.SetDataFormat(FDataFormat);
end;

constructor TGenericMemory.Create(Size: LongInt);
begin
	PMemory := AllocMem(Size+1);
	MemorySize := Size;
end;

destructor TGenericMemory.Destroy;
begin
	FreeMem(PMemory,MemorySize);
end;

function TGenericMemory.ReadByte(Address: Word): Byte;
begin
	if ((Address >= 0) and (Address <= MemorySize + 1)) then
	begin
		RESULT := PMemory[Address];
	end else
	begin
		ShowMessage('Read error  - address ' + IntToHex(Address,4) + ' out of bounds');
		RESULT := $00;
	end;
end;

procedure TGenericMemory.WriteByte(Address: Word; Value: Byte);
begin
	if ((Address >= 0) and (Address <= MemorySize + 1)) then
	begin
		PMemory[Address] := Value;
	end else
	begin
		ShowMessage('Write error - address ' + IntToHex(Address,4) + ' out of bounds');
	end;
end;

//***************************//
//	TMapperA                 //
//***************************//
constructor TMapperA.Create(MAROM, MAVROM, MASRAM: TMemPointer;
	ImageSize: LongInt);
var
	iX:	Integer;
begin
	if ROMHeader.Mapper = 1 then
	begin
		{reset registers}
		for iX := 0 to 3 do
		begin
			M1Regs[iX] := $0C;
		end;
		{reset array}
		for iX := 1 to 5 do
		begin
			M1Array[iX] := 0;
		end;

		{defaults}
		M1Counter := 1;
		M1HighBank := 7;
		M1LowBank := 0;
		DecodedLow := M1LowBank * $4000;
		DecodedHigh := M1HighBank * $4000;
	end;
end;

destructor TMapperA.Destroy;
begin
	{}
end;

function TMapperA.ReadByte(Address: Word): Byte;
begin
	Case ROMHeader.Mapper of
//***************************//
//	Mapper 0                 //
//***************************//
		0:
			begin
      	case Address of
        	$8000..$BFFF:	{low rom}
          	begin
              Result := ROM.ReadByte(Address - $8000);
            end;
          $C000..$FFFF:	{high rom}
          	begin
              if (ROMHeader.PrgSize = $4000) then //	Bottom 16kb has to be mirrored to top 16kb
              begin
                Result := ROM.ReadByte((Address - $8000) and $3fff);
              end else
              begin
                Result := ROM.ReadByte(Address - $8000);
              end;
            end;
          $6000..$7FFF:	{sram}
          	begin
							Result := SRAM.ReadByte(Address - $6000);
            end;
          $0000..$1FFF:	{vrom}
          	begin
            	Result := VROM.ReadByte(Address);
            end;
        else
        	begin
     				Result := $00;	//	None of the above
						ShowMessage('Mapper 0 Read error');
          end;
        end;
      end;
//***************************//
//	Mapper 1                 //
//***************************//
		1:
			begin
					//	ROM read
					if (Address and $8000) > 0 then
					begin
						if (ROMHeader.PrgSize = $4000) then //	Bottom 16kb has to be mirrored to top 16kb
						begin
							Result := ROM.ReadByte((Address - $8000) + DecodedLow);
							//	Read from ROM
							exit;
						end else
						begin
							if (Address >= $C000) then
							begin
								Result := ROM.ReadByte(Integer((Address - $8000) + DecodedHigh));
								//	Read from ROM
								exit;
							end else
							begin
								Result := ROM.ReadByte((Address - $8000) + DecodedLow);
								//	Read from ROM
								exit;
							end;
						end;
					end;
					//	SRAM read
					if (Address and $6000) > 0 then
					begin
						Result := SRAM.ReadByte(Address - $6000);
						//	Read from SRAM
						exit;
					end;
					//	VROM read
					if (Address < $2000) then
					begin
						Result := VROM.ReadByte(Address);
						//	Read from VROM
						exit;
					end;
					ShowMessage('Mapper 1 Read error');
			end;
		else
			Result := $00;
		end;
end;

procedure TMapperA.WriteByte(Address: Word; Value: Byte);
begin
	case ROMHeader.Mapper of
		0:	{default mapper}
			begin
				//frmCMNES.mnuHalt.Click;
				//	Can't write to ROM
				//ShowMessage('Mapper 0 Write error - address '+ IntToHex(Address,4));
			end;
		1:	{mapper 1}
			begin
				if (Value and $80) > 0 then {reset register}
				case Address of
					$8000..$9FFF:
						begin
							M1Regs[0] := $0C; {reset bits 2 & 3 to logic 1}
							mmoEvents.Lines.Add('Register 0 - Reset');
							M1Counter := 1;
							exit;
						end;
					$A000..$BFFF:
						begin
							M1Regs[1] := $0C; {reset bits 2 & 3 to logic 1}
							mmoEvents.Lines.Add('Register 1 - Reset');
							M1Counter := 1;
							exit;
						end;
					$C000..$DFFF:
						begin
							M1Regs[2] := $0C; {reset bits 2 & 3 to logic 1}
							mmoEvents.Lines.Add('Register 2 - Reset');
							M1Counter := 1;
							exit;
						end;
					$E000..$FFFF:
						begin
							M1Regs[3] := $0C; {reset bits 2 & 3 to logic 1}
							mmoEvents.Lines.Add('Register 3 - Reset');
							M1Counter := 1;
							exit;
						end;
					end;

				if (M1Counter <> 5) then
				begin
					if (Value and $01) > 0 then
						M1Array[M1Counter] := $01 shl M1Counter - 1
					else
						M1Array[M1Counter] := $00;
					Inc(M1Counter);
					exit;
				end else
				begin
					M1Counter := 1;
					if (Value and $01) > 0 then
						M1Array[M1Counter] := $01 shl M1Counter - 1
					else
						M1Array[M1Counter] := $00;
				end;

				{register 0}
				if (Address >= $8000) and (Address <= $9FFF) then
				begin
					M1Regs[0] := M1Regs[0] or (M1Array[0]);
					M1Regs[0] := M1Regs[0] or (M1Array[1]);
					M1Regs[0] := M1Regs[0] or (M1Array[2]);
					M1Regs[0] := M1Regs[0] or (M1Array[3]);
					M1Regs[0] := M1Regs[0] or (M1Array[4]);
					//M1Regs[0] := M1Regs[0] shr 1;
					mmoEvents.Lines.Add('Register 0 - Value = ' + IntToHex(M1Regs[0],2));
					exit;
				end;

				{register 1}
				if (Address >= $A000) and (Address <= $BFFF) then
				begin
					M1Regs[1] := M1Regs[1] or (M1Array[0]);
					M1Regs[1] := M1Regs[1] or (M1Array[1]);
					M1Regs[1] := M1Regs[1] or (M1Array[2]);
					M1Regs[1] := M1Regs[1] or (M1Array[3]);
					M1Regs[1] := M1Regs[1] or (M1Array[4]);
					mmoEvents.Lines.Add('Register 1 - Value = ' + IntToHex(M1Regs[1],2));
					exit;
				end;

				{register 2}
				if (Address >= $C000) and (Address <= $DFFF) then
				begin
					M1Regs[2] := M1Regs[2] or (M1Array[0]);
					M1Regs[2] := M1Regs[2] or (M1Array[1]);
					M1Regs[2] := M1Regs[2] or (M1Array[2]);
					M1Regs[2] := M1Regs[2] or (M1Array[3]);
					M1Regs[2] := M1Regs[2] or (M1Array[4]);
					mmoEvents.Lines.Add('Register 2 - Value = ' + IntToHex(M1Regs[2],2));
					exit;
				end;

				{register 3}
				if (Address >= $E000) and (Address <= $FFFF) then
				begin
					M1Regs[3] := M1Regs[3] or (M1Array[1]);
					M1Regs[3] := M1Regs[3] or (M1Array[2]);
					M1Regs[3] := M1Regs[3] or (M1Array[3]);
					M1Regs[3] := M1Regs[3] or (M1Array[4]);

					mmoEvents.Lines.Add('Register 3 - Value = ' + IntToHex(M1Regs[3],2));

					M1Regs[3] := M1Regs[3] shl 5;
					M1Regs[3] := M1Regs[3] shr 5;

					if (M1Regs[0] and $04) > 0 then
					begin {swap $8000}
						if (M1Regs[0] and $08) > 0 then
						begin {16k}
							M1LowBank := M1Regs[3];
							DecodedLow := M1LowBank * $4000;
							//ShowMessage('A');
						end else
						begin {32k}
							M1LowBank := (M1Regs[3]);
							M1HighBank := M1LowBank + 1;
							DecodedLow := ((M1LowBank - 1) * $4000);
							DecodedHigh := ((M1HighBank - 1) * $4000);
							ShowMessage('B');
						end;
					end else
					begin	{swap $C000}
						if (M1Regs[0] and $08) > 0 then
						begin {16k}
							M1HighBank := M1Regs[3];
							DecodedHigh := (M1HighBank - 1 * $4000);
							ShowMessage('C');
						end else
						begin {32k}
							M1LowBank := (M1Regs[3]);
							M1HighBank := M1LowBank + 1;
							DecodedLow := (M1LowBank - 1 * $4000);
							DecodedHigh := (M1HighBank - 1 * $4000);
							ShowMessage('D');
						end;
					end;
					exit;
				end;
			end;
		end;
end;

//***************************//
//	TMemBus                  //
//***************************//
constructor TMemBus.Create(BIRAM: TGenericMemory; BMapper: TMapperA;BPPU: TPPU);
begin
	{}
end;

destructor TMemBus.Destroy;
begin
	{}
end;

function TMemBus.ReadAddress(Address: Word): Word;
var
	tempWord: Word;
	tempByte: Byte;
begin
	tempByte := ReadByte(Address);		//	Read lower half of address
	tempWord := ReadByte(Address+1);	//	Read top half of address
	tempWord  := tempWord shl 8;
	tempWord := tempWord or Word(tempByte) and $00FF;
	Result := tempWord;
end;

function TMemBus.ReadByte(Address: Word): Byte;
begin
	case Address of
		$0000..$1FFF:
			Result := IRAM.ReadByte(Address and $07FF);	//	Read from IRAM
		$2000..$3FFF:
			Result := PPU.ReadByte((Address and $2007));
		$4014:
			Result := PPU.ReadByte(Address);
		$4015:
			Result := $FF;
		$4016:
			Result := GamePad1.ReadByte;
		$4017:
			Result := GamePad2.ReadByte;
		$6000..$FFFF:
			Result := Mapper0.ReadByte(Address);
		else
			begin
				Result := $00;
				ShowMessage('MemBus Read error');
			end;
		end;
end;

procedure TMemBus.WriteByte(Address: Word; Value: Byte);
begin
	case Address of
		$0000..$1FFF:
			IRAM.WriteByte(Address and $07FF,Value);	//	Write to IRAM
		$2000..$3FFF:
			PPU.WriteByte((Address and $2007),Value);
		$4000..$4013:
			exit;
		$4014:
			PPU.WriteByte(Address,Value);
		$4015:
			exit;
		$4016:
			GamePad1.WriteByte(Address,Value);
		$4017:
			GamePad2.WriteByte(Address,Value);
		$4018..$5FFF: {unknown}
			Value := Value;
		$6000..$FFFF:
			Mapper0.WriteByte(Address, Value);				//	Write to Mapper
		else
			begin
				ShowMessage('MemBus Write error' + IntToHex(Address,4) );
				frmCMNES.mnuHalt.Click;
			end;
		end;
end;


//***************************//
//	TCPU                     //
//***************************//
function TCPU.AbsoluteOperandAddress: Word;
begin
	Result := MEMBUS.ReadAddress(PC-2);
end;

function TCPU.AbsoluteXOperandAddress: Word;
begin
// the word following the opcode is added to register X
// to give the address of the operand
	Result := ((MEMBUS.ReadAddress(PC-2) + RegX) and $FFFF);
end;

function TCPU.AbsoluteYOperandAddress: Word;
begin
// the word following the opcode is added to register Y
// to give the address of the operand
	Result := ((MEMBUS.ReadAddress(PC-2) + RegY) and $FFFF);
end;

function TCPU.BitSet(BitNumber: ShortInt): Byte;
begin
	if (SR and ($01 shl BitNumber)) > 0 then
	begin
		Result := $01;
	end else
	begin
		Result := $00;
	end;
end;

constructor TCPU.Create(MemoryBus: TMemBus);
begin
	{}
end;

destructor TCPU.Destroy;
begin
	{}
end;

function TCPU.ExecuteInstruction(var Cycles: Int64): Byte;
var
	tempByte:	Byte;
	tempWord:	Word;
	tempByte2: Byte;
begin

	IR := MEMBUS.ReadByte(PC);

case IR of
	$00..$80:
		begin
			case IR of
				$00:	//	BRK
					begin
						Inc(Cycles,7);
						PushWord(PC);
						PushByte(SR);
						PC := MEMBUS.ReadAddress(IRQ_BRK_VECTOR);
						SRSetBit(SR_INTR_FLAG);
						SRSetBit(SR_BREAK_FLAG);
						{exit;}
					end;
				$01:	//	ORA indirect x - Logical OR of mem and acc
					begin
						Inc(Cycles,6);
						Inc(PC,2);	//	2 byte opcode
						tempWord := IndirectXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc or tempByte;	// perform the ORA operation
						SRSetNZ(RegAcc);	// set flags in status register
						{exit;}
					end;
				$05:	// ORA Z - Logical OR of mem and Acc Zero Page
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc or tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$06:	// ASL Z - Arithmetic Shift Left zero page
					begin
						Inc(Cycles,5);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						if (tempByte and $80) > 0 then // if the bit to be shifted is a one
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						tempByte := tempByte shl 1;
						SRSetNZ(tempByte); 	// set N and Z correctly
						MEMBUS.WriteByte(tempWord,tempByte);
						{exit;}
					end;
				$08:	// PHP - push processor status (Status Reg) on stack
					begin
						Inc(Cycles,3);
						Inc(PC);
						PushByte(SR);
						{exit;}
					end;
				$09:	// ORA #n - Logical Or of Mem and Acc
					begin
						Inc(Cycles,2);
						Inc(PC,2);
						tempWord := ImmediateOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc or tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$0A:	// ASL accumulator
					begin
						Inc(Cycles,2);
						Inc(PC);
						if (RegAcc and $80) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						RegAcc := RegAcc shl 1;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$0D:	// ORA Q - Logical Or of mem and Acc absolute
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc or tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$0E:	// ASL Q - Arithmetic Shift Left absolute
					begin
						Inc(Cycles,6);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						if (tempByte and $80) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);		//	Set the carry bit
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);	//	Clear the carry bit
						end;
						tempByte := tempByte shl 1;
	  				SRSetNZ(tempByte);
						MEMBUS.WriteByte(tempWord, tempByte);
						{exit;}
					end;
				$10:	// BPL - Branch if Plus
					begin
						Inc(PC,2);
						if (BitSet(SR_NEG_FLAG)) > 0 then
						begin							//	N=1
							Inc(Cycles,2);	// only 2 cycles if no branch taken
						end else
						begin							//	N=0
							Inc(Cycles,4);
							PC := RelativeOperandAddress();
						end;
						{exit;}
					end;
				$11:	// ORA indirect Y - Logical Or of mem and acc
					begin
						Inc(Cycles,5);
						Inc(PC,2);
						tempWord := IndirectYOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc or tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$15:	// ORA Z Page, X - Logical Or of mem and Acc Zero Page x
					begin
						Inc(Cycles,4);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc or tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$16: 	// ASL Z,X - arithmetic shift left memory zero page X
					begin
						Inc(Cycles,6);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						if (tempByte and $80) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						tempByte := tempByte shl 1;
						SRSetNZ(tempByte);
						MEMBUS.WriteByte(tempWord,tempByte);
						{exit;}
					end;
				$18:	// CLC - clear carry flag
					begin
						Inc(PC);
						Inc(Cycles,2);
						SRClearBit(SR_CARRY_FLAG);
						{exit;}
					end;
				$19:	// ORA Q,Y- OR Acc absolute Y
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteYOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc or tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$1D:	// ORA Q,X- OR Acc absolute X
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc or tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$1E:	// ASL Q,X - Arithmetic Shift Left absolute,X
					begin
						Inc(Cycles,7);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						if (tempByte and $80) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						tempByte := tempByte shl 1;
						SRSetNZ(tempByte);
						MEMBUS.WriteByte(tempWord,tempByte);
						{exit;}
					end;
				$20:	// JSR - jump to subroutine
					begin
						Inc(Cycles,6);
						Inc(PC,3);
						tempWord := PC - 1;
						PushWord(tempWord);
						tempWord := AbsoluteOperandAddress();
						PC := tempWord;
						{exit;}
					end;
				$21:	// AND (Z,X) - Logical AND of Mem and Acc
					begin
						Inc(Cycles,6);
						Inc(PC,2);
						tempWord := IndirectXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc and tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$24:	// BIT Z - bit test with acc
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						if (tempByte and $80) > 0 then
						begin
							SRSetBit(SR_NEG_FLAG);
						end else
						begin
							SRClearBit(SR_NEG_FLAG);
						end;
						if (tempByte and $40) > 0 then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						if (tempByte and RegAcc) > 0 then	// they are non zero
						begin
							SRClearBit(SR_ZERO_FLAG);
						end else
						begin
							SRSetBit(SR_ZERO_FLAG);
						end;
						{exit;}
					end;
				$25:	// AND Z - Logical AND of mem and Acc Zero Page
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc and tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$26:	// ROL Z - rotate left Zero Page
					begin
						Inc(Cycles,5);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						if (tempByte and $80) > 0 then	//	Save MSB
						begin
							tempByte2 := $01;
						end else
						begin
							tempByte2 := $00;
						end;
						tempByte := tempByte shl 1;
						tempByte := tempByte or (SR and $01);	//	put C flag into LSB
						MEMBUS.WriteByte(tempWord,tempByte);
						if (tempByte2) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						SRSetNZ(tempByte);
						exit
					end;
				$28:	// PLP - pull processor status (SR) off stack
					begin
						Inc(Cycles,4);
						Inc(PC);
						SR := PopByte();
						{exit;}
					end;
				$29:	// AND #n - Logical AND of Mem and Acc
					begin
						Inc(Cycles,2);
						Inc(PC,2);
						tempWord := ImmediateOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc and tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$2A:	// ROL - rotate left accumulator
					begin
						Inc(Cycles,2);
						Inc(PC);
						if (RegAcc and $80) > 0 then // save MSB in temp_byte
						begin
							tempByte := $01;
						end else
						begin
							tempByte := $00;
						end;
						RegAcc := RegAcc shl 1;
						RegAcc := RegAcc or (SR and $01);
						if (tempByte) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$2C:	// BIT Q - bit test with acc
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						SRSetN(tempByte);
						if (tempByte and $40) > 0 then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						SRSetZ(tempByte and RegAcc);
						{exit;}
					end;
				$2D:	// AND Q - Logical AND of mem and Acc absolute
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc and tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$2E:	// ROL - rotate left absolute
					begin
						Inc(Cycles,6);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						if (tempByte and $80) > 0 then
						begin
							tempByte2 := $01;
						end else
						begin
							tempByte2 := $00;
						end;
						tempByte := tempByte shl 1;
						tempByte := tempByte or (SR and $01);
						MEMBUS.WriteByte(tempWord,tempByte);
						if (tempByte2) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						SRSetNZ(tempByte);
						{exit;}
					end;
				$30:	// BMI - branch if minus
					begin
						Inc(PC,2);
						if(BitSet(SR_NEG_FLAG)) > 0 then
						begin								// N=1
							Inc(Cycles,4);
							PC := RelativeOperandAddress();
						end else
						begin								// N=0
							Inc(Cycles,2);
						end;
						{exit;}
					end;
				$31:	// AND (Z),Y - Logical AND of Mem and Acc
					begin
						Inc(Cycles,5);
						Inc(PC,2);
						tempWord := IndirectYOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc and tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$35:	// AND Z,X - load Y zero page X
					begin
						Inc(Cycles,4);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc and tempByte;
						SRSetNZ(RegAcc);
						exit
					end;
				$36:	// ROL Z,X - rotate left Zero Page X
					begin
						Inc(Cycles,6);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						if (tempByte and $80) > 0 then
						begin
							tempByte2 := $01;
						end else
						begin
							tempByte2 := $00;
						end;
						tempByte := (tempByte shl 1);
						tempByte := tempByte or (SR and $01);
						MEMBUS.WriteByte(tempWord,tempByte);
						if (tempByte2) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						SRSetNZ(tempByte);
						{exit;}
					end;
				$38:	// SEC - set carry flag
					begin
						Inc(Cycles,2);
						Inc(PC);
						SRSetBit(SR_CARRY_FLAG);
						{exit;}
					end;
				$39:	// AND Q,Y - Logical AND of mem and Acc absolute, Y
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteYOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc and tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$3D:	// AND Q,X - Logical AND of Mem and Acc
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc and tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$3E:	// ROL Q,X- rotate left absolute X
					begin
						Inc(Cycles,7);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						if (tempByte and $80) > 0 then
						begin
							tempByte2 := $01;
						end else
						begin
							tempByte2 := $00;
						end;
						tempByte := tempByte shl 1;
						tempByte := tempByte or (SR and $01);
						MEMBUS.WriteByte(tempWord,tempByte);
						if (tempByte2) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						SRSetNZ(tempByte);
						{exit;}
					end;
				$40:	// RTI - return from interrupt
					begin
						Inc(Cycles,6);
						Inc(SP);
						SR := MEMBUS.ReadByte(SP);
						Inc(SP,2);
						PC := MEMBUS.ReadAddress(SP-1);
						{exit;}
					end;
				$41:	// EOR (Z,X) - exclusive OR memory indirect X
					begin
						Inc(Cycles,6);
						Inc(PC,2);
						tempWord := IndirectXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc xor tempByte;	// EOR (XOR)
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$45:	// EOR Z - exclusive OR memory zero page
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc xor tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$46:	// LSR Z - logical shift right zero page
					begin
						Inc(Cycles,5);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						if (tempByte and $01) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						tempByte := tempByte shr 1;
						MEMBUS.WriteByte(tempWord,tempByte);
						SRSetNZ(tempByte);
						{exit;}
					end;
				$48:	// PHA - push Acc on stack
					begin
						Inc(Cycles,3);
						Inc(PC);
						PushByte(RegAcc);
						{exit;}
					end;
				$49:	// EOR #n - exclusive OR memory immediate
					begin
						Inc(Cycles,2);
						Inc(PC,2);
						tempWord := ImmediateOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc xor tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$4A:	// LSR accumulator
					begin
						Inc(Cycles,2);
						Inc(PC);
						if (RegAcc and $01) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						RegAcc := RegAcc shr 1;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$4C:	// JMP absolute
					begin
						Inc(Cycles,3);
						Inc(PC,3);
						PC := AbsoluteOperandAddress();
						{exit;}
					end;
				$4D:	// EOR Q - exclusive OR memory absolute
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc xor tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$4E:	// LSR Q - logical shift right absolute
					begin
						Inc(Cycles,6);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						if (tempByte and $01) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						tempByte := tempByte shr 1;
						MEMBUS.WriteByte(tempWord,tempByte);
						SRSetNZ(tempByte);
						{exit;}
					end;
				$50:	// BVC - branch if overflow clear
					begin
						Inc(PC,2);
						if (BitSet(SR_OVERFLOW_FLAG)) > 0 then //	V=1
						begin
							Inc(Cycles,2);
						end else
						begin															// V=0
							Inc(Cycles,4);
							PC := RelativeOperandAddress();
						end;
						{exit;}
					end;
				$51:	// EOR (Z),Y - exclusive OR memory indirect Y
					begin
						Inc(Cycles,5);
						Inc(PC,2);
						tempWord := IndirectYOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc xor tempByte;
						SRSetNZ(RegAcc);
						exit
					end;
				$55:	// EOR Z Page, X - exclusive OR memory zero page X
					begin
						Inc(Cycles,4);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc xor tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$56:	// LSR Z Page, X - logical shift right zero page X
					begin
						Inc(Cycles,6);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress;
						tempByte := MEMBUS.ReadByte(tempWord);
						if (tempByte and $01) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						tempByte := tempByte shr 1;
						MEMBUS.WriteByte(tempWord,tempByte);
						SRSetNZ(tempByte);
						{exit;}
					end;
				$58:	// CLC - clear IRQ disable flag
					begin
						Inc(PC);
						Inc(Cycles,2);
						SRClearBit(SR_INTR_FLAG);
						{exit;}
					end;
				$59:	// EOR Q, Y - exclusive OR memory absolute Y
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteYOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc xor tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$5D:	// EOR Q, X - exclusive OR memory absolute X
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						RegAcc := RegAcc xor tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$5E:	// LSR Q,X - logical shift right absolute X
					begin
						Inc(Cycles,7);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						if (tempByte and $01) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						tempByte := tempByte shr 1;
						MEMBUS.WriteByte(tempWord,tempByte);
						SRSetNZ(tempByte);
						{exit;}
					end;
				$60:	// RTS return to subroutine
					begin
						Inc(Cycles,6);
						PC := MEMBUS.ReadAddress(SP+1)+1;
						Inc(SP,2);
						// no flags affected
						{exit;}
					end;
				$61:  // ADC (Z,X) - add acc w/ carry
					begin
						Inc(Cycles,6);
						Inc(PC,2);
						tempWord := IndirectXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						tempByte2 := RegAcc;
						tempWord := tempByte + RegAcc + 1*BitSet(SR_CARRY_FLAG);
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);
						if (tempWord and $100) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag
						// if A and B have the *SAME* sign then overflow *IS* possible
						//    and if A and A' have different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) = 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$65:	// ADC Z - add acc with carry Zero Page
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						tempByte2 := RegAcc;
						tempWord := tempByte + RegAcc + 1*BitSet(SR_CARRY_FLAG);
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);
						if (tempWord and $100) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						// if A and B have the *SAME* sign then overflow *IS* possible
						//    and if A and A' have different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) = 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$66:	// ROR Z- rotate right zero page
					begin
						Inc(Cycles,5);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						if (tempByte and $01) > 0 then
						begin
							tempByte2 := $80;
						end else
						begin
							tempByte2 := $00;
						end;
						tempByte := tempByte shr 1;
						tempByte := tempByte or ((SR and $01) shl 7);
						MEMBUS.WriteByte(tempWord,tempByte);
						if (tempByte2) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						SRSetNZ(tempByte);
						{exit;}
					end;
				$68:	// PLA - pull Acc off stack
					begin
						Inc(Cycles,4);
						Inc(PC);
						RegAcc := PopByte();
						SRSetNZ(RegAcc);
						{exit;}
					end;

				$69:		// ADC #n - add acc with carry
					begin
						Inc(Cycles,2);
						Inc(PC,2);
						tempWord := ImmediateOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						tempByte2 := RegAcc;		// save original acc
						tempWord := tempByte + RegAcc + 1*BitSet(SR_CARRY_FLAG);	// do the add
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);		// set N and Z flags
						if (tempWord and  $100) > 0	then // Set Carry flag
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag
						// if A and B have the *SAME* sign then overflow *IS* possible
						//    and if A and A' have different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) = 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end	else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$6A:		// ROR - rotate right accumulator
					begin
						Inc(Cycles,2);
						Inc(PC);
						if (RegAcc and $01) > 0	then // save LSB in tempByte
						begin
							tempByte := $80;
						end	else
						begin
							tempByte := $00;
						end;
						RegAcc := RegAcc shr 1;		// shift right 1
						RegAcc := RegAcc or ((SR  and  $01) shl 7);		// put C flag into MSB
						if (tempByte) > 0 then			// put LSB into C flag
						begin
							SRSetBit(SR_CARRY_FLAG);
						end	else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$6C:		// JMP indirect
					begin
						Inc(Cycles,5);
						Inc(PC,3);
						PC := IndirectOperandAddress();
						// no flags affected
						{exit;}
					end;
				$6D:		// ADC Q - add acc with carry absolute
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();		// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						tempByte2 := RegAcc;		// save original acc
						tempWord := tempByte + RegAcc + 1*BitSet(SR_CARRY_FLAG);	// do the add
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);		// set N and Z flags
						if (tempWord and $100) > 0 then	// Set Carry flag
						begin
							SRSetBit(SR_CARRY_FLAG);
						end	else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag
						// if A and B have the *SAME* sign then overflow *IS* possible
						//    and if A and A' have different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) = 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end	else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$6E:		// ROR Q- rotate right absolute
					begin
						Inc(Cycles,6);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();		// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						if (tempByte and $01) > 0 then 	// save LSB in tempByte
						begin
							tempByte2 := $80;
						end	else
						begin
							tempByte2 := $00;
						end;
						tempByte := tempByte shr 1;		// shift right 1
						tempByte := tempByte or ((SR and $01) shl 7);		// put C flag into MSB
						MEMBUS.WriteByte(tempWord, tempByte);
						if (tempByte2) > 0 then			// pub LSB into C flag
						begin
							SRSetBit(SR_CARRY_FLAG);
						end	else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						SRSetNZ(tempByte);
						{exit;}
					end;
				$70:                    // BVS - branch if overflow set
					begin
						Inc(PC,2);
						if (BitSet(SR_OVERFLOW_FLAG)) > 0 then  // V=1
						begin
							Inc(Cycles,4);
							PC := RelativeOperandAddress();
						end	else                            // V=0
						begin
							Inc(Cycles,2);
							// no flags affected
						end;
						{exit;}
					end;
				$71:                  // ADC (Z),Y - add acc w/ carry
					begin
						Inc(Cycles,5);
						Inc(PC,2);
						tempWord := IndirectYOperandAddress();	// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// read data
						tempByte2 := RegAcc;		// save original acc
						tempWord := tempByte + RegAcc + 1*BitSet(SR_CARRY_FLAG);	// do the add
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);		// set N and Z flags
						if (tempWord and $100) > 0 then	// Set Carry flag
						begin
							SRSetBit(SR_CARRY_FLAG);
						end else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag
						// if A and B have the *SAME* sign then overflow *IS* possible
						//    and if A and A' have different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) = 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end	else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$75:                  // ADC Z PAGE, X - add acc w/ carry
					begin
						Inc(Cycles,4);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();	// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// read data
						tempByte2 := RegAcc;		// save original acc
						tempWord := tempByte + RegAcc + 1*BitSet(SR_CARRY_FLAG);	// do the add
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);		// set N and Z flags
						if (tempWord and $100) > 0 then	// Set Carry flag
						begin
							SRSetBit(SR_CARRY_FLAG);
						end	else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag
						// if A and B have the *SAME* sign then overflow *IS* possible
						//    and if A and A' have different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) = 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end	else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$76:		// ROR Z,X - rotate right Zero Page X
					begin
						Inc(Cycles,6);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();		// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						if (tempByte and $01) > 0 then	// save LSB
						begin
							tempByte2 := $80;
						end	else
						begin
							tempByte2 := $00;
						end;
						tempByte := tempByte shr 1;		// shift right 1
						tempByte := tempByte or ((SR and $01) shl 7);		// put C flag into MSB
						MEMBUS.WriteByte(tempWord, tempByte);
						if (tempByte2) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end	else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						SRSetNZ(tempByte);
						{exit;}
					end;
				$78:	// SEI - disable interrupts
					begin
						Inc(PC);
						Inc(Cycles,2);
						SRSetBit(SR_INTR_FLAG);
						{exit;}
					end;
				$79:		// ADC Q,Y - add acc with carry absolute Y
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteYOperandAddress();		// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						tempByte2 := RegAcc;		// save original acc
						tempWord := tempByte + RegAcc + 1*BitSet(SR_CARRY_FLAG);	// do the add
						RegAcc := tempWord  and  $00FF;
						SRSetNZ(RegAcc);		// set N and Z flags
						if (tempWord and $100) > 0 then	// Set Carry flag
						begin
							SRSetBit(SR_CARRY_FLAG);
						end	else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag
						// if A and B have the *SAME* sign then overflow *IS* possible
						//    and if A and A' have different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) = 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end	else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$7D:		// ADC Q,X - add acc with carry absolute X
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();		// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						tempByte2 := RegAcc;		// save original acc
						tempWord := tempByte + RegAcc + 1*BitSet(SR_CARRY_FLAG);	// do the add
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);		// set N and Z flags
						if (tempWord and $100) > 0 then	// Set Carry flag
						begin
							SRSetBit(SR_CARRY_FLAG);
						end	else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag
						// if A and B have the *SAME* sign then overflow *IS* possible
						//    and if A and A' have different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) = 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end	else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$7E:		// ROR Q,X- rotate right absolute X
					begin
						Inc(Cycles,7);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();		// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						if (tempByte and $01)	> 0 then // save LSB in tempByte
						begin
							tempByte2 := $80;
						end	else
						begin
							tempByte2 := $00;
						end;
						tempByte := tempByte shr 1;		// shift right 1
						tempByte := tempByte or ((SR and $01) shl 7);		// put C flag into MSB
						MEMBUS.WriteByte(tempWord, tempByte);
						if (tempByte2) > 0 then
						begin
							SRSetBit(SR_CARRY_FLAG);
						end	else
						begin
							SRClearBit(SR_CARRY_FLAG);
						end;
						SRSetNZ(tempByte);
						{exit;}
					end;
			end;
		end;
	$81..$FF:
		begin
			case IR of

				$81:		// STA (Z, X) - store acc indirect ,X
					begin
						Inc(Cycles,6);
						Inc(PC,2);
						tempWord := IndirectXOperandAddress();		// get address
						MEMBUS.WriteByte(tempWord, RegAcc);
						// no flags affected
						{exit;}
					end;
				$84:		// STY - store Y in memory Zero Page
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();		// get address
						MEMBUS.WriteByte(tempWord, RegY);
						// no flags altered
						{exit;}
					end;
				$85:		// STA Z - Store A in Memory zero page
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();	// read in address
						MEMBUS.WriteByte(tempWord, RegAcc);	// write to memory
						// no flags affected
						{exit;}
					end;
				$86:		// STX Z - Store X in Memory zero page
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();	// read in address
						MEMBUS.WriteByte(tempWord, RegX);	// write to memory
						// no flags affected
						{exit;}
					end;
				$88:		// DEY - decrement Y
					begin
						Inc(Cycles,2);
						Inc(PC);
            asm
							SUB CPU.RegY,1
						end;
						SRSetNZ(RegY);	// set the flags
						{exit;}
					end;
				$8A:		// TXA - transfer X to A
					begin
						Inc(Cycles,2);
						Inc(PC);
						RegAcc := RegX;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$8C:		// STY Q - Store Y in Memory absolute
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();	// read in address
						MEMBUS.WriteByte(tempWord, RegY);	// write to memory
						// no flags affected
						{exit;}
					end;
				$8D:		// STA absolute- store accumulator
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();	// read in address
						MEMBUS.WriteByte(tempWord, RegAcc);	// write to memory
						// no flags affected
						{exit;}
					end;
				$8E:		// STX - Store X in Memory absolute
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();	// read in address
						MEMBUS.WriteByte(tempWord, RegX);	// write to memory
						// no flags affected
						{exit;}
					end;
				$90:		// BCC - branch if carry clear
					begin
						Inc(PC,2);
						if (BitSet(SR_CARRY_FLAG)) > 0 then	// C=1
						begin
							Inc(Cycles,2);
						end	else			// C=0
						begin
							Inc(Cycles,4);
							PC := RelativeOperandAddress();
						end;
						// no flags affected
						{exit;}
					end;

				$91:		// STA (Z),Y - store acc indirect ,Y
					begin
						Inc(Cycles,5);
						Inc(PC,2);
						tempWord := IndirectYOperandAddress();
						MEMBUS.WriteByte(tempWord, RegAcc);
						// no flags affected
						{exit;}
					end;

				$94:		// STY Z,X - store Y zero page X
					begin
						Inc(Cycles,4);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();	// get address
						MEMBUS.WriteByte(tempWord ,RegY);
						// no flags affected
						{exit;}
					end;

				$95:		// STA Z,X - store Acc zero page X
					begin
						Inc(Cycles,4);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();	// get address
						MEMBUS.WriteByte(tempWord ,RegAcc);
						// no flags affected
						{exit;}
					end;

				$96:		// STX Z,Y - store Acc zero page Y
					begin
						Inc(Cycles,4);
						Inc(PC,2);
						tempWord := ZeroPageYOperandAddress();	// get address
						MEMBUS.WriteByte(tempWord ,RegX);
						// no flags affected
						{exit;}
					end;

				$98:		// TYA - transfer Y to Acc
					begin
						Inc(Cycles,2);
						Inc(PC);
						RegAcc := RegY;		// make the transfer
						SRSetNZ(RegAcc);		// set the flags
						{exit;}
					end;

				$99:		// STA Q,Y - store Acc absolute Y
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteYOperandAddress();	// read in address
						MEMBUS.WriteByte(tempWord ,RegAcc);
						// no flags affected
						{exit;}
					end;

				$9A:		// TXS - transfer X to S
					begin
						Inc(Cycles,2);
						Inc(PC);
						SP := RegX;
						SP := SP and $01ff;
						SP := SP or $0100;
						// no flags affected
						{exit;}
					end;

				$9D:		// STA Q,X - store Acc absolute X
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();	// read in address
						MEMBUS.WriteByte(tempWord ,RegAcc);
						// no flags affected
						{exit;}
					end;

				$A0:		// LDY - load Y immediate
					begin
						Inc(Cycles,2);
						Inc(PC,2);
						tempWord := ImmediateOperandAddress();
						RegY := MEMBUS.ReadByte(tempWord);
						SRSetNZ(RegY);
						{exit;}
					end;

				$A1:		// LDA (Z,X) - load Acc indirect X
					begin
						Inc(Cycles,6);
						Inc(PC,2);
						tempWord := IndirectXOperandAddress();
						RegAcc := MEMBUS.ReadByte(tempWord);
						SRSetNZ(RegAcc);
						{exit;}
					end;

				$A2:		// LDX #n - Load X immediate
					begin
						Inc(Cycles,4);
						Inc(PC,2);
						tempWord := ImmediateOperandAddress();
						RegX := MEMBUS.ReadByte(tempWord);		// fetch the immediate value
						SRSetNZ(RegX);		// set the status register flags
						{exit;}
					end;

				$A4:		// LDY Z - load register Y zero page
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();		// get the address
						RegY := MEMBUS.ReadByte(tempWord);
						SRSetNZ(RegY);
						{exit;}
					end;

				$A5:		// LDA - load accumulator zero page
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();		// get the address
						RegAcc := MEMBUS.ReadByte(tempWord);
						SRSetNZ(RegAcc);
						{exit;}
					end;

				$A6:		// LDX - load register X zero page
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();		// get the address
						RegX := MEMBUS.ReadByte(tempWord);
						SRSetNZ(RegX);
						{exit;}
					end;
				$A8:		// TAY - tranfer A to Y
					begin
						Inc(Cycles,2);
						Inc(PC);
						RegY := RegAcc;
						SRSetNZ(RegY);
						{exit;}
					end;
				$A9:		// LDA immediate
					begin
						Inc(Cycles,2);
						Inc(PC,2);
						tempWord := ImmediateOperandAddress();
						RegAcc := MEMBUS.ReadByte(tempWord);	// read in value
						SRSetNZ(RegAcc);				// set appropriate flags in status register
						{exit;}
					end;

				$AA:		// TAX - transfer A to X
					begin
						Inc(Cycles,2);
						Inc(PC);
						RegX := RegAcc;	// make the tranfer
						SRSetNZ(RegX);
						{exit;}
					end;

				$AC:		// LDY Q - load register Y absolute
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();	// read in address
						RegY := MEMBUS.ReadByte(tempWord);
						SRSetNZ(RegY);
						{exit;}
					end;

				$AE:		// Load X absolute
					begin
						Inc(Cycles,4);			// this instruction takes 4 NES clock cycles
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();	// read in address
						RegX := MEMBUS.ReadByte(tempWord);
						SRSetNZ(RegX);			// set the status register flags
						{exit;}
					end;

				$AD:		// Load Accumulator absolute
					begin
						Inc(Cycles,4);			// this instruction takes 4 NES clock cycles
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();	// read in address
						RegAcc := MEMBUS.ReadByte(tempWord);
						SRSetNZ(RegAcc);			// set the status register flags
						{exit;}
					end;

				$B0:		// BCS - branch if carry set
					begin
						Inc(PC,2);
						if (BitSet(SR_CARRY_FLAG)) > 0 then 	// C=1
						begin
							Inc(Cycles,4);		// 4 cycles if branch taken
							PC := RelativeOperandAddress();
						end	else			// Z=0
						begin
							Inc(Cycles,2);
						end;
						// no flags affected
						{exit;}
					end;

				$B1:		// LDA (Z),Y - load acc indirect Y
					begin
						Inc(Cycles,5);
						Inc(PC,2);
						tempWord := IndirectYOperandAddress();
						RegAcc := MEMBUS.ReadByte(tempWord);
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$B4:		// LDY Z,X - load Y zero page X
					begin
						Inc(Cycles,4);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();	// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// read data
						RegY := tempByte;
						SRSetNZ(RegY);
						{exit;}
					end;
				$B5:		// LDA Z,X - load Acc zero page X
					begin
						Inc(Cycles,4);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();	// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// read data
						RegAcc := tempByte;
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$B6:		// LDX Z,Y- load X zero page Y
					begin
						Inc(Cycles,4);
						Inc(PC,2);
						tempWord := ZeroPageYOperandAddress();	// get address
						RegX := MEMBUS.ReadByte(tempWord);	// get data
						SRSetNZ(RegX);
						{exit;}
					end;
				$B8:		// CLV - clear overflow flag
					begin
						Inc(Cycles,2);
						Inc(PC);
						SRClearBit(SR_OVERFLOW_FLAG);		// clear the D bit in SR
						{exit;}
					end;
				$B9:		// LDA Q,Y- load A absolute Y
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteYOperandAddress();
						RegAcc := MEMBUS.ReadByte(tempWord);
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$BA:		// TSX - transfer S to X
					begin
						Inc(Cycles,2);
						Inc(PC);
						RegX := BYTE(SP and $00ff);
						SRSetNZ(RegX);
						{exit;}
					end;
				$BC:		// LDY Q,X- load Y absolute X
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();		// get address
						RegY := MEMBUS.ReadByte(tempWord);	// get data
						SRSetNZ(RegY);
						{exit;}
					end;
				$BD:		// LDA Q,X- load Acc absolute X
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();
						RegAcc := MEMBUS.ReadByte(tempWord);
						SRSetNZ(RegAcc);
						{exit;}
					end;
				$BE:		// LDX Q,Y- load X absolute Y
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteYOperandAddress();		// get address
						RegX := MEMBUS.ReadByte(tempWord);	// get data
						SRSetNZ(RegX);
						{exit;}
					end;
				$C0:		// CPY #n - compare mem and Y immediate
					begin
						Inc(Cycles,2);
						Inc(PC,2);
						tempWord := ImmediateOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);	// read in data
						SetCMPFlags(tempByte, RegY);		// set the flags accordingly
						{exit;}
					end;
				$C1:		// CMP (Z,X) - compare mem and Acc indirect X
					begin
						Inc(Cycles,6);
						Inc(PC,2);
						tempWord := IndirectXOperandAddress();		// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// read data
						SetCMPFlags(tempByte, RegAcc);		// set the flags accordingly
						{exit;}
					end;
				$C4:		// CPY Z - compare mem and Y zero page
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);	// read in data
						SetCMPFlags(tempByte, RegY);		// set the flags accordingly
						{exit;}
					end;
				$C5:		// CMP Z - compare mem and Acc zero page
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();		// read in address
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						SetCMPFlags(tempByte, RegAcc);		// set the flags accordingly
						{exit;}
					end;
				$C6:		// DEC - decrement memory by one Zero Page
					begin
						Inc(Cycles,5);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();		// read in address
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						asm
							SUB tempByte,1
						end;
						MEMBUS.WriteByte(tempWord, tempByte);	// put it back
						SRSetNZ(tempByte);		// set flags
						{exit;}
					end;
				$C8:		// INY - increment Y
					begin
						Inc(PC);
						Inc(Cycles,2);
						Inc(RegY);		// increment
						SRSetNZ(RegY);
						{exit;}
					end;
				$C9:		// CMP #n - compare mem and Acc Absolute
					begin
						Inc(Cycles,2);
						Inc(PC,2);
						tempWord := ImmediateOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);	// read in data
						SetCMPFlags(tempByte, RegAcc);		// set the flags accordingly
						{exit;}
					end;
				$CA:		// DEX - decrement X
					begin
						Inc(Cycles,2);
						Inc(PC);
						asm
							SUB CPU.RegX,1
						end;
						SRSetNZ(RegX);	// set the flags
						{exit;}
					end;
				$CC:		// CPY Q - compare mem and Y absolute
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);	// read in data
						SetCMPFlags(tempByte, RegY);		// set the flags accordingly
						{exit;}
					end;
				$CD:		// CMP Q - compare mem and Acc Absolute
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();	// get address from opcode
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						SetCMPFlags(tempByte, RegAcc);		// set the flags accordingly
						{exit;}
					end;
				$CE:		// DEC Q - Decrement memory absolute
					begin
						Inc(Cycles,6);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();	// get address from opcode
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						asm
							SUB tempByte,1
						end;
						MEMBUS.WriteByte(tempWord, tempByte);	// store it back
						SRSetNZ(tempByte);
						{exit;}
					end;
				$D0:		// BNE - branch if not equal
					begin
						Inc(PC,2);
						if (BitSet(SR_ZERO_FLAG)) > 0 then 	// Z=1
						begin
							Inc(Cycles,2);
						end else			// Z=0
						begin
							Inc(Cycles,4);		// 4 cycles if branch taken
							PC := RelativeOperandAddress();
						end;
						// no flags affected
						{exit;}
					end;
				$D1:		// CMP (Z), Y - compare mem and Acc indirect Y
					begin
						Inc(Cycles,5);
						Inc(PC,2);
						tempWord := IndirectYOperandAddress();	// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// read data
						SetCMPFlags(tempByte, RegAcc);		// set the flags accordingly
						{exit;}
					end;
				$D5:		// CMP Z PAGE, X - compare mem and Acc Absolute
					begin
						Inc(Cycles,4);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();	// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// read data
						SetCMPFlags(tempByte, RegAcc);		// set the flags accordingly
						{exit;}
					end;
				$D6:		// DEC Z,X - decrement memory zero page X
					begin
						Inc(Cycles,6);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						asm
							SUB tempByte,1
						end;
						MEMBUS.WriteByte(tempWord, tempByte);	// write it back
						SRSetNZ(tempByte);
						{exit;}
					end;
				$D8:		// CLD - clear decimal mode
					begin
						Inc(PC);
						Inc(Cycles,2);
						SRClearBit(SR_DECIMAL_FLAG);		// clear the D bit in SR
						{exit;}
					end;
				$D9:		// CMP Q,Y - compare mem and Acc Absolute Y
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteYOperandAddress();	// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						SetCMPFlags(tempByte, RegAcc);		// set the flags accordingly
						{exit;}
					end;
				$DD:		// CMP Q,X - compare mem and Acc Absolute X
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();	// get address
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						SetCMPFlags(tempByte, RegAcc);		// set the flags accordingly
						{exit;}
					end;
				$DE:		// DEC Q,X - decrement memory absolute X
					begin
						Inc(Cycles,7);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();	// get address
						tempByte := MEMBUS.ReadByte(tempWord);
						asm
							SUB tempByte,1
						end;
						MEMBUS.WriteByte(tempWord, tempByte);	// write it back
						SRSetNZ(tempByte);
						{exit;}
					end;
				$E0:		// CPX #n - compare mem and X immediate
					begin
						Inc(Cycles,2);
						Inc(PC,2);
						tempWord := ImmediateOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);	// read in data
						SetCMPFlags(tempByte, RegX);		// set the flags accordingly
						{exit;}
					end;
				$E1:		// SBC (Z,X) - Subtract acc with carry indirect X
					begin
						Inc(Cycles,6);
						Inc(PC,2);
						tempWord := IndirectXOperandAddress();		// get address
						tempByte := MEMBUS.ReadByte(tempWord);
						tempByte2 := RegAcc;		//save the original Acc
						tempWord := RegAcc - tempByte  - (1-1*BitSet(SR_CARRY_FLAG));	// do the sub
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);		// set N and Z flags
						if (tempWord and $100) > 0 then	// Set Carry flag
						begin
							SRClearBit(SR_CARRY_FLAG);
						end	else
						begin
							SRSetBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag:
						// if A and B are different signs then overflow possible
						// and if A and A' are different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) > 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$E4:		// CPX Z - compare mem and X zero page
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						SetCMPFlags(tempByte, RegX);		// set the flags accordingly
						{exit;}
					end;
				$E5:		// SBC z - Subtract acc with carry zero page
					begin
						Inc(Cycles,3);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						tempByte2 := RegAcc;		//save the original Acc
						tempWord := RegAcc - tempByte  - (1-1*BitSet(SR_CARRY_FLAG));	// do the add
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);		// set N and Z flags
						if (tempWord and $100) > 0 then	// Set Carry flag
						begin
							SRClearBit(SR_CARRY_FLAG);
						end	else
						begin
							SRSetBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag:
						// if A and B are different signs then overflow possible
						// and if A and A' are different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) > 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end	else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$E6:		// INC Z - increment memory by one Zero Page
					begin
						Inc(Cycles,5);
						Inc(PC,2);
						tempWord := ZeroPageOperandAddress();		// read in address
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						Inc(tempByte);		// increment it
						MEMBUS.WriteByte(tempWord, tempByte);	// put it back
						SRSetNZ(tempByte);		// set flags
						{exit;}
					end;
				$E8:		// INX - increment X
					begin
						Inc(PC);
						Inc(Cycles,2);
						asm
							ADD CPU.RegX,1
						end;
						SRSetNZ(RegX);
						{exit;}
					end;
				$E9:		// SBC #n - Subtract acc with carry
					begin
						Inc(Cycles,2);
						Inc(PC,2);
						tempWord := ImmediateOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						tempByte2 := RegAcc;		//save the original Acc
						tempWord := RegAcc - tempByte  - (1-1*BitSet(SR_CARRY_FLAG));	// do the add
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);		// set N and Z flags
						if (tempWord and $100) > 0 then // Set Carry flag
						begin
							SRClearBit(SR_CARRY_FLAG);
						end	else
						begin
							SRSetBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag:
						// if A and B are different signs then overflow possible
						// and if A and A' are different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) > 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end	else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$EA:		// NOP - no operation
					begin
						Inc(Cycles,2);
						Inc(PC);
						// not hard to implement :)
						{exit;}
					end;
				$EC:		// CPX Q - compare mem and X absolute
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);	// read in data
						SetCMPFlags(tempByte, RegX);		// set the flags accordingly
						{exit;}
					end;
				$ED:		// SBC Q - Subtract acc with carry Abolute
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);	// get data
						tempByte2 := RegAcc;		//save the original Acc
						tempWord := RegAcc - tempByte  - (1-1*BitSet(SR_CARRY_FLAG));	// do the add
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);		// set N and Z flags
						if (tempWord and $100) > 0 then 	// Set Carry flag
						begin
							SRClearBit(SR_CARRY_FLAG);
						end	else
						begin
							SRSetBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag:
						// if A and B are different signs then overflow possible
						// and if A and A' are different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) > 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end	else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$EE:		// INC Q - increment memory by one Zero Page
					begin
						Inc(Cycles,6);
						Inc(PC,3);
						tempWord := AbsoluteOperandAddress();	// get address from opcode			;
						tempByte := MEMBUS.ReadByte(tempWord);
						Inc(tempByte);		// increment it
						MEMBUS.WriteByte(tempWord, tempByte);	// put it back
						SRSetNZ(tempByte);		// set flags
						{exit;}
					end;
				$F0:		// BEQ - branch if equal
					begin
						Inc(PC,2);
						if (BitSet(SR_ZERO_FLAG)) > 0 then	// Z=1
						begin
							Inc(Cycles,4);		// 4 cycles if branch taken
							PC := RelativeOperandAddress();
						end else
						begin
							Inc(Cycles,2);
						end;
						// no flags affected
						{exit;}
					end;
				$F1:		// SBC (Z),Y - Subtract acc with carry indirect Y
					begin
						Inc(Cycles,5);
						Inc(PC,2);
						tempWord := IndirectYOperandAddress();		// get address
						tempByte := MEMBUS.ReadByte(tempWord);
						tempByte2 := RegAcc;		//save the original Acc
						tempWord := RegAcc - tempByte  - (1-1*BitSet(SR_CARRY_FLAG));	// do the sub
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);		// set N and Z flags
						if (tempWord and $100) > 0 then	// Set Carry flag
						begin
							SRClearBit(SR_CARRY_FLAG);
						end	else
						begin
							SRSetBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag:
						// if A and B are different signs then overflow possible
						// and if A and A' are different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) > 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$F5:		// SBC Z,X - Subtract acc with carry zero page, X
					begin
						Inc(Cycles,4);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();	// get address
						tempByte := MEMBUS.ReadByte(tempWord);
						tempByte2 := RegAcc;		//save the original Acc
						tempWord := RegAcc - tempByte  - (1-1*BitSet(SR_CARRY_FLAG));	// do the sub
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);		// set N and Z flags
						if (tempWord and $100) > 0 then	// Set Carry flag
						begin
							SRClearBit(SR_CARRY_FLAG);
						end	else
						begin
							SRSetBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag:
						// if A and B are different signs then overflow possible
						// and if A and A' are different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) > 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end	else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$F6:		// INC Z,X - increment memory zero page X
					begin
						Inc(Cycles,6);
						Inc(PC,2);
						tempWord := ZeroPageXOperandAddress();
						tempByte := MEMBUS.ReadByte(tempWord);
						Inc(tempByte);			// do the increment
						MEMBUS.WriteByte(tempWord, tempByte);	// write it back
						SRSetNZ(tempByte);
						{exit;}
					end;
				$F8:                  // SED - set decimal flag
					begin
						Inc(Cycles,2);
						Inc(PC,1);
						SRSetBit(SR_DECIMAL_FLAG);	 // set the decimal flag
						{exit;}
					end;
				$F9:		// SBC Q,Y - Subtract acc with carry
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteYOperandAddress();	// get address
						tempByte := MEMBUS.ReadByte(tempWord);
						tempByte2 := RegAcc;		//save the original Acc
						tempWord := RegAcc - tempByte  - (1-1*BitSet(SR_CARRY_FLAG));	// do the sub
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);		// set N and Z flags
						if (tempWord and $100) > 0 then	// Set Carry flag
						begin
							SRClearBit(SR_CARRY_FLAG);
						end	else
						begin
							SRSetBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag:
						// if A and B are different signs then overflow possible
						// and if A and A' are different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) > 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end	else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$FD:		// SBC Q,X - Subtract acc with carry absolute, X
					begin
						Inc(Cycles,4);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();	// get address
						tempByte := MEMBUS.ReadByte(tempWord);
						tempByte2 := RegAcc;		//save the original Acc
						tempWord := RegAcc - tempByte  - (1-1*BitSet(SR_CARRY_FLAG));	// do the sub
						RegAcc := tempWord and $00FF;
						SRSetNZ(RegAcc);		// set N and Z flags
						if (tempWord and $100) > 0 then	// Set Carry flag
						begin
							SRClearBit(SR_CARRY_FLAG);
						end	else
						begin
							SRSetBit(SR_CARRY_FLAG);
						end;
						// now set overflow flag:
						// if A and B are different signs then overflow possible
						// and if A and A' are different signs then overflow occured
						if (((tempByte2 xor tempByte) and $80) > 0) and
							 (((tempByte2 xor RegAcc) and $80) > 0) then
						begin
							SRSetBit(SR_OVERFLOW_FLAG);
						end	else
						begin
							SRClearBit(SR_OVERFLOW_FLAG);
						end;
						{exit;}
					end;
				$FE:		// INC Q,X - increment memory absolute X
					begin
						Inc(Cycles,7);
						Inc(PC,3);
						tempWord := AbsoluteXOperandAddress();	// get address
						tempByte := MEMBUS.ReadByte(tempWord);
						Inc(tempByte);			// do the increment
						MEMBUS.WriteByte(tempWord, tempByte);	// write it back
						SRSetNZ(tempByte);
						{exit;}
					end;
				end;
			end;
	else
		begin
			ShowMessage('Unsupported opcode encountered ' + IntToHex(IR,2) + ' at ' + IntToHex(PC,4));
			Inc(PC);
			frmCMNES.mnuHalt.Click;
		end;
	end;
end;


function TCPU.ImmediateOperandAddress: Word;
begin
	Result := (PC-1);		// Check this!
end;

function TCPU.IndirectOperandAddress: Word;
begin
	Result := MEMBUS.ReadAddress(MEMBUS.ReadAddress(PC-2));
end;

function TCPU.IndirectXOperandAddress: Word;
begin
// the byte following the opcode is added to register X to give
// the address on page 0 of the operand
	Result := (MEMBUS.ReadAddress((MEMBUS.ReadByte(PC-1) + RegX) and $00FF));

end;

function TCPU.IndirectYOperandAddress: Word;
begin
// the byte following the opcode is added to register Y to give
// the address on page 0 of the operand
	Result := (MEMBUS.ReadAddress(MEMBUS.ReadByte(PC-1) and $00FF) + RegY);
end;

function TCPU.PopByte: Byte;
begin
// Pops a byte off of the stack, and returns it
	Inc(SP); 			// increment Stack pointer
	Result := MEMBUS.ReadByte(SP);
end;

(*function TCPU.PopWord: Word;
begin
// Pops a word off of the stack, and returns it.
	Inc(SP,2);			// increment Stack pointer
	Result := MEMBUS.ReadAddress(SP-1);
end;*)

procedure TCPU.PushByte(Value: Byte);
begin
// pushes byte onto the stack
	MEMBUS.WriteByte(SP,Value); // put it on the stack
	Dec(SP);    											// update the stack pointer
end;

procedure TCPU.PushWord(Value: Word);
var
	tempByte:	Byte;
begin
// pushes word onto the stack
	tempByte := (Value shr 8);				// get the high byte
	PushByte(tempByte);
	tempByte := (Value and $00FF);				// get the low byte
	PushByte(tempByte);
end;

function TCPU.RelativeOperandAddress: Word;
begin
// the byte following the opcode is added to the PC to give the
// address of the operand (the byte value is added via 2's complement,
// this means if the number is negative, we sign extend it)

	if (MEMBUS.ReadByte(PC-1) and $80) > 0 then
	begin
		Result := PC + ((MEMBUS.ReadByte(PC-1) or $FF00));
	end else
	begin
		Result := PC + ((MEMBUS.ReadByte(PC-1) and $00FF));
	end;

end;

procedure TCPU.Reset;
begin
	ScanLine := 0;
	CycleCount := 0;
	TotalCycles := 0;
	SP := Word($01FF); 		// reset stack pointer to top of stack
	SR := Byte($20);			// status register get's cleared except bit 5
	RegAcc := 0;					// clear registers
	RegX := 0;
	RegY := 0;

	// now retrieve the correct starting value of PC from the RESET_VECTOR
	PC := MEMBUS.ReadAddress(RESET_VECTOR);

	mmoEvents.Lines.Add('CPU - Reset');

end;

procedure TCPU.SetBus(MemoryBus: TMemBus);
begin
	{}
end;

procedure TCPU.SetCMPFlags(Value1, Value2: Byte);
var
	tempByte:	Byte;
begin
// sets the status register flags accordingly
// value2 = regAcc, should do value2 - value1

	tempByte := Value2 - Value1;		//	compare
	SRSetNZ(tempByte);							//	set N and Z

	if (Value2 >= Value1) then
	begin
		SRSetBit(SR_CARRY_FLAG);
	end else
	begin
		SRClearBit(SR_CARRY_FLAG);
	end;

end;

procedure TCPU.SRClearBit(Position: Integer);
begin
 //	Clear bit
	SR := SR and not(1 shl(Position));
end;

procedure TCPU.SRSetBit(Position: Integer);
begin
//	Set bit
	SR := SR or (1 shl(Position));
end;

procedure TCPU.SRSetN(Value: Byte);
begin
// Set the N flag in SR based on the byte value passed
	SR := (SR and $7F) or ((Value) and $80);
end;

procedure TCPU.SRSetNZ(Value: Byte);
begin
//	Set N and Z flags based on the byte value passed
	if (Value and $80) > 0 then
	begin
		SRSetBit(SR_NEG_FLAG);
	end else
	begin
		SRClearBit(SR_NEG_FLAG);
	end;
	if Value = $00 then
	begin
		SRSetBit(SR_ZERO_FLAG)
	end	else
	begin
		SRClearBit(SR_ZERO_FLAG);
	end;
end;

procedure TCPU.SRSetZ(Value: Byte);
begin
// Set the Z flag in SR based on the byte value passed
	{SR := (SR and $FD) or (not(Value) ? $02 : $00);  }  // Check this
	if (Value <> 0) then
	begin
		SRClearBit(SR_ZERO_FLAG);
	end else
	begin
		SRSetBit(SR_ZERO_FLAG);
	end;

end;

procedure TCPU.VBlankInterrupt(Cycles: Int64);
begin
// execute a vblank interrupt
	Inc(CycleCount,7);
	PushWord(PC); // push the Program counter on stack
	PushByte(SR);	// push the status register on stack
	PC := MEMBUS.ReadAddress(NMI_VECTOR);
end;

function TCPU.ZeroPageOperandAddress: Word;
begin
// the byte following the opcode is the address on page 0
// retrieve the byte at that address
	Result := (MEMBUS.ReadByte(PC-1) and $00FF);
end;

function TCPU.ZeroPageXOperandAddress: Word;
begin
// the byte following the opcode is added to register X
// to give the address on page 0 of the operand
	Result := ((MEMBUS.ReadByte(PC-1) + RegX) and $00FF);
end;

function TCPU.ZeroPageYOperandAddress: Word;
begin
// the byte following the opcode is added to register Y
// to give the address on page 0 of the operand
	Result := ((MEMBUS.ReadByte(PC-1) + RegY) and $00FF);
end;

//***************************//
//	TTile                    //
//***************************//
constructor TTile.Create;
begin
	{}
end;

destructor TTile.Destroy;
begin
	{}
end;

function TTile.GetColor(Row, Col: Byte): Byte;
begin
	Result := Colour[Row,Col];
end;

procedure TTile.SetColour(Row, Col, Value: Byte);
begin
	if (Row < 9) and (Col < 9) then
		Colour[Row,Col] := Value;
end;

//***************************//
//	TPPU                     //
//***************************//
constructor TPPU.Create(PASRAM: TGenericMemory; Mirroring: Byte;
	PASMapper: TMapperA);
var
	i:		Integer;
begin
	PPUMirroring := Mirroring;
	BackScrollHorizontal := 0;
	BackScrollVertical := 0;
	VRAMAddrReg := 0;
	VRAMAddrRegBuff := 0;
	UpdateBackground := 1;
	PtnTableUpdated := 1;
	VRAMDataBuff := $FF;
	BackgroundColor := 13;

  PVRAM := VRAM.PMemory;

	{reset PPU registers}
	for i := 0 to 8 do
	begin
		PPURegs[i] := 0;
	end;

  {reset sprite palette}
	for i := 0 to 16 do
	begin
		SpritePalette[i] := BackgroundColor;
	end;

  {reset image palette}
	for i := 0 to 16 do
	begin
		ImagePalette[i] := BackgroundColor;
	end;

end;

destructor TPPU.Destroy;
begin
	{}
end;

procedure TPPU.DisplayPalette;
var
	i:		Integer;
	x:		Integer;
	y:		Integer;
begin
	NESScreen.LockBackground(NESScreen.g_pDDSBack);
	for i := 0 to 16 do
	begin
		for x := 0 to 10 do
		begin
			for y := 0 to 10 do
			begin
				NESScreen.SetPixel(NESScreen.g_pDDSBack, i*10+x, y+30, SpritePalette[i]);	// sprite palette
				NESScreen.SetPixel(NESScreen.g_pDDSBack, i*10+x, y, ImagePalette[i]);	// image palette
			end;
		end;
	end;
	NESScreen.UnlockSurface(NESScreen.g_pDDSBack);
end;

procedure TPPU.DisplayPatternTable;
var
	a,b,c,d:  		Integer;
const
	NUMROWS = 16;
	NUMCOLS = 32;
begin

	NESScreen.LockBackground(NESScreen.g_pDDSBack);
	FillPatternTableCache();

	for a := 0 to NUMROWS do
	begin
		for b := 0 to NUMCOLS do
		begin
			for c := 0 to 7 do
			begin
				for d := 0 to 7 do
				begin
					if (b < $0F) then
					begin
						NESScreen.SetPixel(NESScreen.g_pDDSBack,(b shl 3)+d, (a shl 3)+c, ImagePalette[MyTableLow[a*numrows+b].GetColor(d,c)]);
					end else
					begin
						NESScreen.SetPixel(NESScreen.g_pDDSBack,(b shl 3)+d, (a shl 3)+c, ImagePalette[MyTableHigh[a*numrows+b].GetColor(d,c)]);
					end;
				end;
			end;
		end;
	end;

	NESScreen.UnlockSurface(NESScreen.g_pDDSBack);
end;

procedure TPPU.Draw8x16Sprite(Y, X, TileIndex, Attributes: Byte);
begin
	{optimizated}
	if (TileIndex and $01) = 0 then {even}
	begin
		if (Attributes and $80) > 0 then
		begin
			{if vertical flip flag set}
			DrawSprite(Y+8, X, TileIndex, Attributes, $0000);		// draw the top half of the sprite
			DrawSprite(Y, X, TileIndex + 1, Attributes, $0000);	// draw the bottom half
		end else
		begin
			DrawSprite(Y, X, TileIndex, Attributes, $0000);		// draw the top half of the sprite
			DrawSprite(Y+8, X, TileIndex + 1, Attributes, $0000);	// draw the bottom half
		end;
	end else	{odd}
	begin
		if (Attributes and $80) > 0 then
		begin
			{if vertical flip flag set}
			DrawSprite(Y, X, TileIndex-1, Attributes, $1000);		// draw the top half of the sprite
			DrawSprite(Y+8, X, TileIndex, Attributes, $1000);	// draw the bottom half
		end else
		begin
			DrawSprite(Y+8, X, TileIndex, Attributes, $1000);		// draw the top half of the sprite
			DrawSprite(Y, X, TileIndex-1, Attributes, $1000);	// draw the bottom half
		end;
	end;

end;

procedure TPPU.DrawBackground;
var
	NameTableBaseAddr:	Word;
	StartCol:						Byte;
	TileRow, TileCol:		Byte;
	TileRowX32:					Word;
	Quad3Address:				Word;
	Quad0Address:				Word;
	Quad1Address:				Word;
	Quad3Num:						Byte;
	Quad0Num:						Byte;
	Quad1Num:						Byte;
	Quad2Num:						Byte;
const
	NUM_ROWS = 29;
	NUM_COLS = 31;
begin

	PPU.BackgroundColor := SpritePalette[0];

	NESScreen.ClearSurface(0);

	{clear variables just to be safe}
	TileRow						:= 0;
	TileRowX32				:= 0;

	NameTableBaseAddr := ( Word(PPURegs[0]) shl 10 );	// get name table from PPU ctrl reg

	asm
		AND NameTableBaseAddr,$0FFF
		OR NameTableBaseAddr,$2000
    MOV BX,NameTableBaseAddr
		MOV AX,BX
		XOR AX,$0400
		MOV quad3Address,AX
		AND AX,$0C00
		SHR AX,10
		MOV quad3Num,AL
		MOV AX,BX
		XOR AX,$0800
		MOV quad0Address,AX
		AND AX,$0C00
		SHR AX,10
		MOV quad0Num,AL
		MOV AX,BX
		XOR AX,$0C00
		MOV quad1Address,AX
		AND AX,$0C00
		SHR AX,10
		MOV quad1Num,AL
		MOV AX,BX
		AND AX,$0C00
		SHR AX,10
		MOV quad2Num,AL
	end;
  {original code}
	(*NameTableBaseAddr := NameTableBaseAddr and $0fff;		// clear out top of address
	NameTableBaseAddr := NameTableBaseAddr or $2000;		// add the 0x200 in
	quad3Address := (NameTableBaseAddr xor $0400);	// base address of name table to be drawn in quadrant 3
	quad3Num := Byte(((quad3Address and $0c00) shr 10));		// name table number to be drawn in quadrant 3

	quad0Address := (NameTableBaseAddr xor $0800);
	quad0Num := Byte(((quad0Address and $0c00) shr 10));

	quad1Address := (NameTableBaseAddr xor $0c00);
	quad1Num := Byte(((quad1Address and $0c00) shr 10));
	quad2Num := Byte(((NameTableBaseAddr and $0c00) shr 10));*)

	NESScreen.LockNESBackground(NESScreen.g_pDDSNESBack);

	if (PPURegs[1] and $02) > 0 then
			StartCol := 0
		else
			StartCol := 1;

	//	Pattern table layout
	//  -------------------
	//  |        |        |
	//  |   2    |   3    |
	//  -------------------
	//  |        |        |
	//  |   0    |   1    |
	//  -------------------

	// now draw the background
	if bBackgroundEnabled then
	begin
		if (PPURegs[1] and $08) > 0 then	// if background is enabled, draw it
		begin
			for TileRow := 0 to NUM_ROWS do
			begin
				//TileRowX32 := TileRow shl 5;
				asm
					MOV AH,0
					MOV AL,TileRow
					SHL AX,5
					MOV TileRowX32,AX
				end;
				for TileCol := StartCol to NUM_COLS do
				begin
					if (TablesToUpdate and $01) > 0 then
					begin
						DrawTile(0, quad0Num,TileRow, TileCol, VRAMReadByte(Quad0Address + TileRowX32 + TileCol));
						bTable0Done := True;
					end;
					if (TablesToUpdate and $02) > 0 then
					begin
						DrawTile(1, quad1Num,TileRow, TileCol, VRAMReadByte(Quad1Address + TileRowX32 + TileCol));
						bTable1Done := True;
					end;
				 	if (TablesToUpdate and $04) > 0 then
					begin
						DrawTile(2, quad2Num,TileRow, TileCol, VRAMReadByte(NameTableBaseAddr + TileRowX32 + TileCol));
						bTable2Done := True;
					end;
					if (TablesToUpdate and $08) > 0 then
					begin
						DrawTile(3, quad3Num,TileRow, TileCol, VRAMReadByte(Quad3Address + TileRowX32 + TileCol));
						bTable3Done := True;
					end;
				end;
			end;
		end else
		begin
			NESScreen.ClearSurface(0);{clear nes background}
		end;
	end;
	UpdateBackground := $00;		// indicate that the background has been updated

	NESScreen.UnLockSurface(NESScreen.g_pDDSNESBack);

end;

procedure TPPU.DrawFrame;
var
	PtnTableAddr: 	Word;
	i:							Integer;
begin
	bSkipFrame := False;
	Inc(iFrameSkip);

	case iSkip of
		1: {25%}
			begin
				if iFrameSkip = 4 then
				begin
					iFrameSkip := 0;
					asm
						INC FrameCount
					end;
					bSkipFrame := True;
				end;
			end;
		2:
			begin {33%}
				if iFrameSkip = 3 then
				begin
					iFrameSkip := 0;
					asm
						INC FrameCount
					end;
					bSkipFrame := True;
				end;
			end;
		3: {50%}
			begin
				if iFrameSkip = 2 then
				begin
					iFrameSkip := 0;
					asm
						INC FrameCount
					end;
					bSkipFrame := True;
				end;
			end;
		4: {75}
			begin
				if iFrameSkip > 1 then
				begin
					if iFrameSkip = 4 then
						iFrameSkip := 0;
					asm
						INC FrameCount
					end;
					bSkipFrame := True;
				end;
			end;
    else
    	asm
    		INC FrameCount
    	end;
		end;


	if bSkipFrame then
		exit;

	//PtnTableAddr := 0;
	i := 0;

// if pattern table has been updated, need to reload the ptn_table_cache, then redraw screen

	TablesToUpdate := $00;

		if ((BackScrollHorizontal <> 255) and (BackScrollVertical <> 239)) then
		begin
    	TablesToUpdate := TablesToUpdate or $04;
		end;

		if ((BackScrollHorizontal <> 0) and (BackScrollVertical <> 239)) then
		begin
    	TablesToUpdate := TablesToUpdate or $08;
		end;

		if ((BackScrollHorizontal <> 255) and (BackScrollVertical <> 0)) then
		begin
    	TablesToUpdate := TablesToUpdate or $01;
		end;

		if ((BackScrollHorizontal <> 0) and (BackScrollVertical <> 0)) then
		begin
    	TablesToUpdate := TablesToUpdate or $02;
		end;

		if (UpdateBackground = 0) then{not changed}
		begin
			if ((TablesToUpdate and $01) > 0) and not bTable0Done then
			begin
				UpdateBackground := $01;
			end;
			if ((TablesToUpdate and $02) > 0) and not bTable1Done then
			begin
				UpdateBackground := $01;
			end;
			if ((TablesToUpdate and $04) > 0) and not bTable2Done then
			begin
				UpdateBackground := $01;
			end;
			if ((TablesToUpdate and $08) > 0) and not bTable3Done then
			begin
				UpdateBackground := $01;
			end;
		end;

		if (PtnTableUpdated <> 0) then	// only do this when pattern table has been updated
		begin
			FillPatternTableCache();		// no need to do this every time
			DrawBackground();
		end else
		begin
			if (UpdateBackground <> 0) then // if name or attribute tables changed, we need to redraw screen, using cached pattern table
			begin
				DrawBackground();
			end;
		end;

	// if nothing has been updated, we don't need to updated the background buffer
	NESScreen.BlitNESBackground(BackScrollHorizontal, BackScrollVertical);		// draw the background

	// now it's time to draw the sprites on top of the background

	NESScreen.LockBackground(NESScreen.g_pDDSBack);

	{hit flag}

	if (PPURegs[1] and $10) > 0 then		// if sprites are enabled, draw them
	begin
		if (PPURegs[0] and $20) > 0 then	// if 8x16 sprites are enabled
		begin
			i := 0;
			while i <= 255 do
      begin
      	if (SpriteRam[i] <> 0) or
           (SpriteRam[i+1] <> 0) or
           (SpriteRam[i+2] <> 0) or
					 (SpriteRam[i+3] <> 0) then
        	Draw8x16Sprite(SpriteRam[i], SpriteRam[i+3], SpriteRam[i+1], SpriteRam[i+2]);
        asm
          ADD i,4
        end;
      end;
		end	else	// draw 8x8 sprites
		begin
			PtnTableAddr := ( ( Word(PPURegs[0]) and $0008) shl 9);	// get pattern table from PPU ctrl reg
			i := 0;
			while i <= 255 do
			begin
      	if (SpriteRam[i] <> 0) or
           (SpriteRam[i+1] <> 0) or
           (SpriteRam[i+2] <> 0) or
					 (SpriteRam[i+3] <> 0) then
					DrawSprite(SpriteRam[i], SpriteRam[i+3], SpriteRam[i+1], SpriteRam[i+2], PtnTableAddr);
				asm
					ADD i,4
				end;
			end;
		end;
	end;
	NESScreen.UnLockSurface(NESScreen.g_pDDSBack);
end;

procedure TPPU.DrawSprite(Y, X, TileIndex, Attributes: Byte; PtnTableAddr: Word);
var
	BitField:			Byte;
	tempByte:			Byte;
	tempByte2:  	Byte;
	tempAddr:			Word;
	xOffset:			Integer;
	yOffset:			Integer;
	AttribColor:	Byte;
	ColorIndex:		Byte;
	c,d:					Byte;
begin

	if bSpritesEnabled then
	begin
		//AttribColor := ((Attributes shl 2) and $0C);
		asm
			MOV AL,Attributes
			SHL AX,2
			AND AL,$0C
			MOV AttribColor,AL
		end;

		for c := 0 to 7 do
		begin
			asm
				MOV BitField,$80
				MOV AX,PtnTableAddr
				MOV BL,TileIndex
				MOV BH,0
				SHL BX,4
				ADD BL,C
				ADD AX,BX
				MOV tempAddr,AX
			end;
			//BitField := $80;		// start with MSB set
			//tempAddr := PtnTableAddr + Word(TileIndex shl 4)+c;
			tempByte := Mapper0.ReadByte(tempAddr);
			tempByte2 := Mapper0.ReadByte(tempAddr + 8);

			for d := 0 to 7 do
			begin
				ColorIndex := 0;
				if (tempByte and BitField) > 0 then
				begin
					ColorIndex := $01;
				end;
				if (tempByte2 and BitField) > 0 then
				begin
					//ColorIndex := ColorIndex or $02;
					asm
						OR ColorIndex,$02
					end;
				end;

				//ColorIndex := ((((tempByte2 and BitField) and $01) shl 1) or (((tempByte and BitField) and $01)));
				asm
					MOV AL,AttribColor
					OR ColorIndex,AL
				end;
				//ColorIndex := ColorIndex or AttribColor;		// factor in the attribute table color index
				//if (ColorIndex > 16) then
				//	ShowMessage('Bad stuff in draw_sprite!');
				//BitField := BitField shr 1;
				asm
					SHR BitField,1
				end;

				xOffset := d;
				yOffset := c+1; {corrects sprite position problem}

				if (Attributes and $40) > 0 then	// if bit 6 set then flip horizontally (draw a vertical line down middle)
					xOffset := 7-d;
				if (Attributes and $80) > 0 then	// if bit 7 set then flip vertically
					yOffset := 7-c;
				if (Y <= 232) then		// don't draw the sprite if it's off the screen (in vblank area of scanlines 240 through 262)
				begin
					if (x + xOffset > 8) or ((PPURegs[1] and $04) > 0) then
					begin
						if ((SpritePalette[ColorIndex] <> ImagePalette[0]) and
               (SpritePalette[ColorIndex] <> BackgroundColor))
								then  // then the pixel is not transparent
						begin
							if (Attributes and $20) > 0 then	// then draw behind background
							begin
								NESScreen.SetPixelBehindBackground( X + XOffset, Y + YOffset, SpritePalette[ColorIndex], SpritePalette[0]);
							end else	// draw it over the background
							begin
								NESScreen.SetPixel(NESScreen.g_pDDSBack, X + XOffset, Y + YOffset, SpritePalette[ColorIndex]);
							end;
						end;
					end;
				end
			end;
		end;
	end;

end;

procedure TPPU.DrawTile(QuadrantNum: Word; TableNum, TileRow: Byte; TileColumn: Word;
	TileIndex: Byte);
var
	PtnTableAddr:			Word;
	ColorIndex:				Byte;
	AttribColor:			Byte;
	Column:						Word;
	Row:							Word;
	c,d:							Integer;
begin
	PtnTableAddr := PPURegs[0];
	asm
		AND PtnTableAddr,$0010
		SHL PtnTableAddr,8
	end;
	//PtnTableAddr := ((Word(PPURegs[0]) and $0010) shl 8);	// get pattern table from PPU ctrl reg

	AttribColor := GetAttribColor(TableNum, TileRow, TileColumn);

	//Column := (TileColumn shl 3) + ((QuadrantNum and $01) shl 8);
	asm
		MOV AX,TileColumn
		//MOV AH,0
		SHL AX,3
		MOV Column,AX
		MOV AX,QuadrantNum
		//MOV AL,0
		AND AX,$01
		SHL AX,8
		ADD Column,AX
	end;
	if (QuadrantNum and $02) > 0 then
	begin
		Row := (TileRow shl 3);
	end else
	begin
		Row := (TileRow shl 3) + 240;
	end;

	begin
		if (PtnTableAddr = $0000) then
		begin
			for c := 0 to 7 do
			begin
				for d := 0 to 7 do
				begin
					ColorIndex := MyTableLow[Tileindex].GetColor(d,c);
					ColorIndex := ColorIndex or AttribColor;		// factor in the attribute table color index
					if ImagePalette[ColorIndex] <> ImagePalette[0] then
						NESScreen.BackSetPixel(NESScreen.g_pDDSNESBack,Column + d, Row + c, ImagePalette[ColorIndex]);
				end;
			end;
		end else
		begin
			for c := 0 to 7 do
			begin
				for d := 0 to 7 do
				begin
					ColorIndex := MyTableHigh[TileIndex].GetColor(d,c);
					ColorIndex := ColorIndex or AttribColor;		// factor in the attribute table color index
					if ImagePalette[ColorIndex] <> ImagePalette[0] then
						NESScreen.BackSetPixel(NESScreen.g_pDDSNESBack,Column + d, Row + c, ImagePalette[ColorIndex]);
				end;
			end;
		end;
	end;

end;

procedure TPPU.EndVBlank;
begin
	PPURegs[2] := PPURegs[2] and $80;	// clear bit 7 because vblank is over
end;

procedure TPPU.FillPatternTableCache;
var
	tempByte:		Byte;
	tempByte2:	Byte;
	tempAddr:		Word;
	BitField:		Byte;
	ColorIndex:	Byte;
	aX,cX,dX:		Byte;
const
	NUMROWS = 16;
	NUMCOLS = 32;
begin

	//BitField := $80;		// start with MSB set
	// pattern table at 0x0000
	for aX := 0 to 255 do
	begin
		for cX := 0 to 7 do
		begin
			BitField := $80;		// start with MSB set
			tempAddr :=  (aX shl 4) + cX;	// index*16 + c
			tempByte := Mapper0.ReadByte(tempAddr);
			tempByte2 := Mapper0.ReadByte(tempAddr + 8);
			for dX := 0 to 7 do
			begin
				ColorIndex := $00;
				if (tempByte and BitField) > 0 then
				begin
					ColorIndex := $02;
				end;
				if (tempByte2 and BitField) > 0 then
				begin
					ColorIndex := ColorIndex or $01;
				end;
				BitField := BitField shr 1;
				MyTableLow[aX].SetColour(dX, cX, ColorIndex);
			end;
		end;
	end;

	// pattern table at 0x1000
	for aX := 0 to 255 do
	begin
		for cX := 0 to 7 do
		begin
			BitField := $80;		// start with MSB set
			tempAddr :=  (aX shl 4) + cX + $1000;	// index*16 + c
			tempByte := Mapper0.ReadByte(tempAddr);
			tempByte2 := Mapper0.ReadByte(tempAddr + 8);
			for dX := 0 to 7 do
			begin
				ColorIndex := $00;
				if (tempByte2 and BitField) > 0 then
				begin
					ColorIndex := $02;
				end;
				if (tempByte and BitField) > 0 then
				begin
					ColorIndex := ColorIndex or $01;
				end;
				BitField  := BitField shr 1;
				MyTableHigh[aX].SetColour(dX,cX,ColorIndex);
			end;
		end;
	end;

	PtnTableUpdated := 0;		// indicate that the pattern table is current

end;

Function TPPU.GetAttribColor(TableNum: Byte;Row: Byte;Col: Byte): Byte;
var
	AttribTableRow:		Byte;
	AttribTableCol:		Byte;
	AttribTableAddr:	Word;
	Color:						Byte;
	BitMask:					Byte;
	NumShifts:				Byte;
begin

	asm
		MOV BitMask,$03
		MOV NumShifts,$00
		MOV AH,0
		MOV AL,TableNum
		SHL AX,10
		ADD AX,$23C0
		MOV AttribTableAddr,AX
		MOV AL,Col
		SHR AL,2
		MOV AttribTableCol,AL
		MOV AL,Row
		SHR AL,2
		MOV AttribTableRow,AL
	end;
(*
	BitMask := $03;
	NumShifts := 0;

	AttribTableAddr := $23c0 + ($400 * TableNum);

	AttribTableCol := Col shr 2;	// column divided by 4
	AttribTableRow := Row shr 2;    *)

	// read in the correct byte from the attribute table
	Color := VRAMReadByte(AttribTableAddr + (AttribTableRow shl 3) + AttribTableCol);

	asm
		MOV AL,Row
		AND AL,$02
		MOV NumShifts,AL
		MOV AL,Col
		AND AL,$02
		MOV AH,0
		SHR AX,1
		OR NumShifts,AL
		SHL NumShifts,1
		MOV CL,NumShifts
		SHR BitMask,CL
		SHR Color,CL
		SHL Color,2
		AND Color,$0C
	end;
	//NumShifts := (Row and $02);
	//NumShifts := NumShifts or ((Col and $02) shr 1);
	//NumShifts := NumShifts shl 1;		// mult by 2
	//BitMask := BitMask shl NumShifts;
	//Color := Color shr NumShifts;
	//Color := Color shl 2;
	//Color := Color and $0c;

	Result := Color;

end;

function TPPU.ReadByte(Address: Word): Byte;
var
	tempByte:	Byte;
	tempWord: Word;
	Index:		Integer;
begin
	//tempByte := 0;
	//tempWord := 0;
	Case Address of
		$2000:
			begin
				Result := PPURegs[0];
				//exit;
			end;
		$2001:
			begin
				Result := PPURegs[1];
				//exit;
			end;
		$2002:
			begin
				tempByte := PPURegs[2];
				PPURegs[2] := PPURegs[2] and $7F; // clear bit 7 after a read from this reg
				PPURegs[5] := 0;	// reset 0x2005 after a read from this reg
				Result := tempByte;
				//exit;
			end;
		$2003:                // sprite ram address register
			begin
				Result := PPURegs[3];
				//exit;
			end;
		$2004:                // sprite ram IO register
			begin
				Index := PPURegs[3];
				Inc(PPURegs[3]);
				(*if (Index < 0) then
				begin
					Index := Index * - 1; {invert}
				end;*)
				Result := SpriteRam[Index]; // Check this
				//exit;
			end;
		$2005:
			begin
				Result := PPURegs[5];
				//exit;
			end;
		$2006:
			begin
				Result := PPURegs[6];
				//exit;
			end;
		$2007:
			begin
				tempWord := VRAMAddrReg;
				if (PPURegs[0] and $04) > 0	then // then bit 2 is set
				begin
					Inc(VRAMAddrReg,32);
				end else
				begin
					Inc(VRAMAddrReg);
				end;
				VRAMAddrReg := VRAMAddrReg and $3FFF; // make sure the address wraps at 0x4000
				if ((tempWord >= $3F00) and (tempWord < $4000)) then
				begin
					if (VRAMAddrReg and $0010) > 0 then
					begin
						Result := SpritePalette[tempWord and $000F];
					end else
					begin
						Result := ImagePalette[tempWord and $000F];
					end;
				end else
				begin
					tempByte := VRAMDataBuff;
					VRAMDataBuff := VRAMReadByte(tempWord);
					Result := tempByte;
				end;
				//exit;
			end;
		else
			ShowMessage('PPU Read error');
			Result := 0;
			exit
		end;

end;

procedure TPPU.SetValues(PASRAM, PASVRAM: TGenericMemory; Mirroring: Byte;
	PASMapper: TMapperA);
var
	i:		Integer;
begin
	PPUMirroring := Mirroring;
	BackScrollHorizontal := 0;
	BackScrollVertical := 0;
	PtnTableUpdated := $01;
	UpdateBackground := $01;
	VRAMDataBuff := $FF;

	//	Reset all PPU registers
	for i := 0 to 8 do
	begin
		PPURegs[i] := 0;
	end;

end;

procedure TPPU.StartVBlank;
begin
	PPURegs[2] := PPURegs[2] and $40; {clear hit}
	PPURegs[2] := PPURegs[2] or $80;  // set bit #7 to indicate we are in vblank
	DrawFrame();	//PPU not Screen
	NESScreen.DrawFrame; //Screen
end;

function TPPU.VRAMReadByte(Address: Word): Byte;
begin
	if (PPUMirroring) > 0 then  // then it is vertical mirroring
	begin
		//Result := VRAM.ReadByte(Address and $07FF); // read from vram device
    Result := PVRAM[Address and $07FF];
	end else	// horizontal mirroring
	begin
		if (Address >= $2800) then
		begin
			//Result := VRAM.ReadByte((Address and $07FF) or $0400);  	// read from vram device
      Result := PVRAM[((Address and $07FF) or $0400)];
		end else
		begin
			//Result := VRAM.ReadByte(Address and $0BFF);	// read from vram device
      Result := PVRAM[Address and $0BFF];
		end;
	end;
end;

procedure TPPU.VRAMWriteByte(Address: Word; Value: Byte);
begin
	if (PPUMirroring) > 0 then	// then it is vertical mirroring
	begin
		//VRAM.WriteByte((Address and $07FF),Value); // read from vram device
    PVRAM[(Address and $07FF)] := Value;
	end else	// horizontal mirroring
	begin
		if (Address >= $2800) then 	// it's to tables 2 or 3
		begin
			//VRAM.WriteByte((Address and $07FF) or $0400, Value);
      PVRAM[((Address and $07FF) or $0400)] := Value;
		end else
		begin
			//VRAM.WriteByte((Address and $0BFF), Value);	// read from vram device
      PVRAM[(Address and $0BFF)] := Value;
		end;
	end;
end;

procedure TPPU.WriteByte(Address: Word; Value: Byte);
var
	i:				Integer;
	tempWord:	Word;
	Index:		Integer;
begin
	//tempWord := 0;
	Case Address of
		$2000:
			begin
				if ((Value and $13) xor (PPURegs[0] and $13)) > 0 then
				begin
					UpdateBackground := $01;
				end;
				PPURegs[0] := Value;
			end;
		$2001:
			begin
				if ((Value and $08) xor (PPURegs[1] and $08)) > 0 then
				begin
					UpdateBackground := $01;
				end;
				PPURegs[1] := Value;
				if (PPURegs[1] and $40) > 0 then {green}
				begin
					BackgroundColor := 25;
					exit;
				end;
				if (PPURegs[1] and $80) > 0 then {blue}
				begin
					BackgroundColor := 34;
					exit;
				end;
				if (PPURegs[1] and $20) > 0 then {red}
				begin
					BackgroundColor := 6;
					exit;
				end;
				BackgroundColor := 13;

				//exit;
				end;
		$2002:
			begin
				PPURegs[2] := Value;
				//exit;
			end;
		$2003:
			begin
				PPURegs[3] := Value;
				//exit;
			end;
		$2004:
			begin
				Index := PPURegs[3];
				(*if Index < 0 then
				begin
					Index := Index * - 1; {invert}
				end;*)
				SpriteRam[Index] := Value;
				PPURegs[3] := PPURegs[3] + 1;
//				exit;
			end;
		$2005:
			begin
				if (PPURegs[5] = Byte($00)) then
				begin
					BackScrollHorizontal := Value;
					PPURegs[5] := $01;
				end else
				begin
					if (Value < 240) then	// ignore if greater than 239
					begin
						BackScrollVertical := Value;
					end;
					PPURegs[5] := $00;
				end;
				//exit;
			end;
		$2006:  // VRAM address register
			begin
				if (PPURegs[6] = Byte($00)) then
				begin
					VRAMAddrRegBuff := Value and $3F;
					VRAMAddrRegBuff := VRAMAddrRegBuff shl 8;
					PPURegs[6] := $01;
				end else
				begin
					VRAMAddrReg := (VRAMAddrRegBuff or Value);
					VRAMAddrRegBuff := $0000;
					PPURegs[6] := $00;
				end;
				//exit;
			end;
		$2007:   // VRAM IO register
			begin
				PPURegs[7] := Value;
				if ((VRAMAddrReg  >= $2000) and (VRAMAddrReg < $3000)) then
				begin
					VRAMWriteByte(VRAMAddrReg, Value);
					UpdateBackground := $01;
				end
				else if ((VramAddrReg >= $3f00) and (VramAddrReg < $4000)) then
				begin
					if (VramAddrReg and $10) > 0 then
					begin
						if (VramAddrReg and $03) > 0 then
						begin
							SpritePalette[VramAddrReg and $0f] := value;
						end else
						begin
							SpritePalette[$00] := Value;
							SpritePalette[$04] := Value;
							SpritePalette[$08] := Value;
							SpritePalette[$0C] := Value;
						end;
					end else
					begin
						if (VramAddrReg and $03) > 0 then
						begin
							ImagePalette[VramAddrReg and $0f] := value;
						end else
						begin
							ImagePalette[$00] := Value;
							ImagePalette[$04] := Value;
							ImagePalette[$08] := Value;
							ImagePalette[$0C] := Value;
						 end;
					end;
					UpdateBackground := $01;		// tell ppu that background needs to be redrawn
				end
				else if (VramAddrReg < $2000) then//  then it is a write to the pattern table
				begin
					{fix me}
					//VROM.WriteByte(VramAddrReg and $07FF, value);
					PtnTableUpdated := $01;
				end
				else
				begin
					//ShowMessage('Invalid VRAM write address!');
				end;
				if (PpuRegs[0] and $04) > 0 then		// then bit 2 is set
				begin
					VramAddrReg := VramAddrReg + 32;
				end	else
				begin
					Inc(VramAddrReg);
				end;
				VramAddrReg  := VramAddrReg and $3fff;		// make sure it wraps around 0x4000
				//exit;
			end;
		$4014: // sprite ram DMA register
			begin
				tempWord := Word(Value) shl 8;
				for i := 0 to 255 do
				begin
					SpriteRam[i] := IRAM.ReadByte(tempWord + i);
				end;
				//exit;
			end;
		else
			ShowMessage('PPU Write error');
		end;
end;

//***************************//
//	TfrmCMNES                //
//***************************//
procedure TfrmCMNES.mnuExitClick(Sender: TObject);
begin
//  NESScreen.Destroy;
	if not bExit then {halt first}
	begin
		bExit := True;
		Application.ProcessMessages;
	end;

  if not mnuOpenROM.Enabled then
	  DestroyNES;

  if NESScreen.WinHandle <> 0 then
  begin
		NESScreen.Destroy;
	  Input.Destroy;
  end;
	Close;
end;

procedure TfrmCMNES.mnuMemoryClick(Sender: TObject);
begin
	frmMemory.Visible := True;
	frmMemory.Show;
end;

procedure TfrmCMNES.mnuOpenROMClick(Sender: TObject);
begin
	dlgOpenROM.InitialDir := strROMPath + '\';
	if dlgOpenROM.Execute then
	begin
		strROMFileName := dlgOpenROM.FileName;
		if ReadROMHeader then	{read ROM header}
		begin
 			{enable / disable menu options}
 			mnuOpenROM.Enabled := False;
			mnuCloseROM.Enabled := True;
			mnuMemory.Enabled := True;
			mnuGo.Enabled := True;
			mnuHalt.Enabled := False;
			mnuClearScreen.Enabled := True;
			mnuCPUReset.Enabled := True;
			mnuDisplayPatternTable.Enabled := True;
			mnuDisplayPalette.Enabled := True;
			mnuROMInfo.Enabled := True;
			mnuDebug.Enabled := True;
			mnuWindowed.Enabled := False;
      {create the NES components}
			CreateNES;
      {hide logo}
			imgTitle.Visible := False;
      {update form caption}
			frmCMNES.Caption := 'CMNES - ' + strROMFileName + ' (loaded)';
			{auto run}
			frmCMNES.mnuGo.Click;
		end else
		begin
      {enable / disable menu options}
			mnuOpenROM.Enabled := True;
			mnuCloseROM.Enabled := False;
			mnuMemory.Enabled := False;
			mnuGo.Enabled := False;
			mnuHalt.Enabled := False;
			mnuClearScreen.Enabled := False;
			mnuCPUReset.Enabled := False;
			mnuDisplayPatternTable.Enabled := False;
			mnuDisplayPalette.Enabled := False;
			mnuROMInfo.Enabled := False;
			mnuDebug.Enabled := False;
      {show logo}
      imgTitle.Visible := True;
      {error}
			ShowMessage('Error opening ' + strROMFileName);
		end;
	end;
end;

function TfrmCMNES.ReadROMHeader: Boolean;
var
	ReadHeader:			TByteArray;
	ROMFile:				TFileStream;
begin

	Result := False;

	ROMFILE := nil;

	try
//  Open ROM
		try
			ROMFile := TFileStream.Create(strROMFileName, fmOpenRead);
		except
			ShowMessage('ROM could not be found!');
			exit;
		end;
		mmoEvents.Lines.Add('ROM Image - Opened');

//	Read ROM
		try
			if ROMFile.Size > 0 then
			begin
				ROMFile.Read(ReadHeader[0],16);
			end;
		except
			ShowMessage('ROM header could not be read!');
			exit;
		end;
		mmoEvents.Lines.Add('ROM Header - Read OK');

		if not (ReadHeader[0] = Ord('N')) or
			 not (ReadHeader[1] = Ord('E')) or
			 not (ReadHeader[2] = Ord('S')) or
			 not (ReadHeader[3] = $1A) then
		begin
			ShowMessage('Not a valid NES file!');
			exit;
		end;

//	Store header
		ROMHeader.PrgSize := (ReadHeader[4]) * $4000;					//	16k
		mmoEvents.Lines.Add('ROM PrgSize = $' + IntToHex(ROMHeader.PrgSize,4));
		if (ReadHeader[5] = 0) then
		begin
			ROMHeader.ChrSize := $2000;
		end else
		begin
			ROMHeader.ChrSize := (ReadHeader[5]) * $2000; 			//  8k
		end;
		mmoEvents.Lines.Add('ROM ChrSize = $' + IntToHex(ROMHeader.ChrSize,4));
		ROMHeader.Mapper := ReadHeader[6];
		ROMHeader.Mapper := ROMHeader.Mapper shr 4;
		ROMHeader.Mapper := ROMHeader.Mapper and $0F;
		mmoEvents.Lines.Add('ROM Mapper = ' + IntToStr(ROMHeader.Mapper));
		ReadHeader[7] := ReadHeader[7] and $F0;
		ROMHeader.Mapper := ROMHeader.Mapper or ReadHeader[7];

		if (ReadHeader[6] and $02) > 0 then						//	Battery backup
		begin
			ROMHeader.BatteryBackup := $01;
			mmoEvents.Lines.Add('ROM Battery Backed = YES');
		end else
		begin
			ROMHeader.BatteryBackup := $00;
			mmoEvents.Lines.Add('ROM Battery Backed = NO');
		end;

		if (ReadHeader[6] and $04) > 0 then						//	Trainer
		begin
			ROMHeader.Trainer := $01;
			mmoEvents.Lines.Add('ROM Trainer = YES');
		end else
		begin
			ROMHeader.Trainer := $00;
			mmoEvents.Lines.Add('ROM Trainer = NO');
		end;

		if (ReadHeader[6] and $08) > 0 then						//	Four screen mirroring
		begin
			ROMHeader.FourScreenMirror := $01;
			ShowMessage('Unsupported four screen mirroring detected!');
		end else
		begin
			ROMHeader.FourScreenMirror := $00;
		end;

		if (ReadHeader[6] and $01) > 0 then						//	Mirroring
		begin
			ROMHeader.Mirroring := $01;									//	Vertical mirroring
			mmoEvents.Lines.Add('ROM Mirroring = VERTICAL');
		end else
		begin
			ROMHeader.Mirroring := $00;									//	Horizonal mirroring
			mmoEvents.Lines.Add('ROM Mirroring = HORIZONTAL');
		end;

		Result := True;

	finally
//	Close file
		ROMFile.Free;
	end;

end;

function TfrmCMNES.ReadROM(PrgSize: Integer; ROM: TGenericMemory;
	ChrSize: Integer; VROM: TGenericMemory; Trainer: Byte): Boolean;
var
	ROMFile:				TFileStream;
	i:							Integer;
	inByte:					Byte;
begin
	RESULT := False;
	ROMFILE := nil;

	try
//  Open ROM
		try
			ROMFile := TFileStream.Create(strROMFileName, fmOpenRead);
		except
			ShowMessage('ROM could not be found!');
			exit;
		end;
//	Read ROM into ROM and VROM objects
		ROMFile.Seek(16,0);					//	Seek past header
		if ROMHeader.Trainer > 0 then
		begin
			ROMFile.Seek(512,16);			//	Seek past trainer
		end;

//	ROM
		try
			for i := 0 to (PrgSize-1) do
			begin
				ROMFile.Read(inByte,1);
				ROM.WriteByte(i,inByte);	//	Read byte into ROM
			end;
		except
			mmoEvents.Lines.Add('Error: ROM load failed');
			exit;
		end;

		mmoEvents.Lines.Add('ROM Loaded OK');

//	VROM
		try
			//if i > 0 then
			//begin
				for i := 0 to ChrSize-1 do
				begin
					ROMFile.Read(inByte,1);
					VROM.WriteByte(i,inByte);	//	Read byte into VROM
				end;
			//end;
		except
			mmoEvents.Lines.Add('Error:	VROM load failed');
			exit;
		end;

		mmoEvents.Lines.Add('VROM Loaded OK');

//	127 byte trailer

	finally
//	Close ROM
		ROMFile.Free;
		mmoEvents.Lines.Add('ROM Image - Closed');
	end;

end;


procedure TfrmCMNES.CreateNES;
begin

	ROM.Create(ROMHeader.PrgSize);				//	Create ROM
	VROM.Create(ROMHeader.ChrSize);				//	Create VROM
	SRAM.Create($2000);										//	SRAM
	IRAM.Create($800);										//	IRAM
	VRAM.Create($800);										//	VRAM

	ReadROM(ROM.MemorySize,ROM,VROM.MemorySize,VROM,0);	//	Read ROM
	if (ROMHeader.Mapper <> 0) and
		 (ROMHeader.Mapper <> 1) then
	begin
		ShowMessage('Mapper ' + IntToStr(ROMHeader.Mapper) + ' not supported!');
		mmoEvents.Lines.Add('Mapper ' + IntToStr(ROMHeader.Mapper) + ' not supported!');
		ROMHeader.Mapper := 0;
	end;
	Mapper0.Create(ROM.PMemory,VROM.PMemory,SRAM.PMemory,ROMHeader.PrgSize);

	ScanLine := 0;
	CycleCount := 0;

	MEMBUS.Create(IRAM,Mapper0,PPU);					//	MEMBUS

	//	CPU
	CPU.Create(MEMBUS);
	mmoEvents.Lines.Add('NES - Created');
	CPU.Reset;

	//	PPU
	PPU.Create(IRAM,ROMHeader.Mirroring,Mapper0);
	PPU.SetValues(IRAM,VRAM,ROMHeader.Mirroring,Mapper0);
	PPU.FillPatternTableCache;


	if NESScreen.WinHandle = 0 then {once only}
		Input.Create;

	GamePad1.Create;
	GamePad1.SetPad(1);

	GamePad2.Create;
	GamePad2.SetPad(2);

	//	pAPU
	//	pAPU.Initialise;

	if NESScreen.WinHandle = 0 then {once only}
	begin
    NESScreen.Initialise(frmCMNES.Handle); 					{create screen}
	end;

	NESScreen.InitPalette;

  NESScreen.ClearSurface(1);											{clear screen}
  NESScreen.ClearSurface(0);
  NESScreen.DrawFrame;														{draw cleared screen}
end;

procedure TfrmCMNES.mnuROMInfoClick(Sender: TObject);
begin
	with TfrmROMInfo.Create(Application) do
	try
		ShowModal
	finally
		free;
	end;

end;

procedure TfrmCMNES.mnuCloseROMClick(Sender: TObject);
begin
	{destroy NES components}
	DestroyNES;
  {enable / disable menu options}
	mnuOpenROM.Enabled := True;
	mnuCloseROM.Enabled := False;
	mnuMemory.Enabled := False;
	mnuGo.Enabled := False;
	mnuHalt.Enabled := False;
	mnuROMInfo.Enabled := False;
	mnuDebug.Enabled := False;
	mnuCPUReset.Enabled := False;
	mnuDisplayPatternTable.Enabled := False;
	mnuClearScreen.Enabled := False;
	mnuDisplayPalette.Enabled := False;
	mnuWindowed.Enabled := True;
	{show logo}
 	imgTitle.Visible := True;
end;

procedure TfrmCMNES.DestroyNES;
begin
	ROM.Destroy;
	VROM.Destroy;
	IRAM.Destroy;
	SRAM.Destroy;
	Mapper0.Destroy;
	MEMBUS.Destroy;
	CPU.Destroy;
	PPU.Destroy;
	VRAM.Destroy;
	mmoEvents.Lines.Add('NES - Destroyed');
end;

procedure TfrmCMNES.mnuGoClick(Sender: TObject);
var
	ProcessID : DWORD;
	ProcessHandle : THandle;
	ThreadHandle : THandle;
begin

	{increase thread priority}
	ProcessID := GetCurrentProcessID;
	ProcessHandle := OpenProcess(PROCESS_SET_INFORMATION,
															 false,
															 ProcessID);
	SetPriorityClass(ProcessHandle, REALTIME_PRIORITY_CLASS);
	ThreadHandle := GetCurrentThread;
	SetThreadPriority(ThreadHandle, THREAD_PRIORITY_HIGHEST);

  {set frame skip}
	SetFrameSkip;

  {enable / disable menu options}
	mnuHalt.Enabled := True;
	mnuGo.Enabled := False;
	mnuOpenROM.Enabled := False;
	mnuCloseROM.Enabled := False;
	mnuDisplayPatternTable.Enabled := False;
	mnuDisplayPalette.Enabled := False;
  mnuSettings.Enabled := False;

	{update form caption}
	frmCMNES.Caption := 'CMNES - ' + strROMFileName + ' (running)';
	mmoEvents.Lines.Add('CPU - Running');

	{enable joystick polling}

	frmCMNES.tmrDIPoll.Enabled := True;

  {main loop}
	while not bExit do
	begin
		CPU.ExecuteInstruction(CycleCount);
		TotalCycles := TotalCycles + CycleCount;
		if TotalCycles > 1000000 then
		begin
			Application.ProcessMessages;
			TotalCycles := 0;
		end;

		if PPU.SpriteRam[0] = ScanLine then
			PPU.PPURegs[2] := PPU.PPURegs[2] or $40; {sprite hit}

		if (CycleCount >= HBLANK) then
		begin
			CycleCount := CycleCount - HBLANK;
			Inc(ScanLine);
			if (ScanLine = SCANLINES) then		// we did the last one already
			begin
				PPU.StartVBlank();	// a vblank should be performed
				if (PPU.ReadByte($2000) and $80) > 0 then
				begin
					CPU.VBlankInterrupt(CycleCount);	// execute an interrupt
				end;
			end;
			if (ScanLine = SCANLINES + VPERIOD/HBLANK) then
			begin
				ScanLine := 0;			// go back to top of screen.
				PPU.EndVBlank();	// end of vblank
			end;
		end;
    {update debug}
		if frmDebug.Visible then
		begin
			UpdateDebug;
		end;
	end;

	bExit := False;

end;

procedure TfrmCMNES.mnuHaltClick(Sender: TObject);
begin
	{log}
	mmoEvents.Lines.Add('CPU - Halted');

  {enable / disable menu options}
	mnuOpenROM.Enabled := False;
	mnuCloseROM.Enabled := True;
	mnuHalt.Enabled := False;
	mnuGo.Enabled := True;
	mnuDisplayPatternTable.Enabled := True;
	mnuDisplayPalette.Enabled := True;
  mnuSettings.Enabled := True;

  {update form caption}
	frmCMNES.Caption := 'CMNES - ' + strROMFileName + ' (halted)';

	{disable joystick polling}
	frmCMNES.tmrDIPoll.Enabled := False;

  {halt main loop}
	bExit := True;
end;

//***************************//
//	TGraphics                //
//***************************//
procedure TfrmCMNES.DXError(HR: HRESULT);
var
	ErrorMessage:		String;
begin
	{directx error message routine}
	case HR of
		DDERR_ALREADYINITIALIZED:           ErrorMessage := 'DDERR_ALREADYINITIALIZED';
		DDERR_CANNOTATTACHSURFACE:          ErrorMessage := 'DDERR_CANNOTATTACHSURFACE';
		DDERR_CANNOTDETACHSURFACE:          ErrorMessage := 'DDERR_CANNOTDETACHSURFACE';
		DDERR_CURRENTLYNOTAVAIL:            ErrorMessage := 'DDERR_CURRENTLYNOTAVAIL';
		DDERR_EXCEPTION:                    ErrorMessage := 'DDERR_EXCEPTION';
		DDERR_GENERIC:                      ErrorMessage := 'DDERR_GENERIC';
		DDERR_HEIGHTALIGN:                  ErrorMessage := 'DDERR_HEIGHTALIGN';
		DDERR_INCOMPATIBLEPRIMARY:          ErrorMessage := 'DDERR_INCOMPATIBLEPRIMARY';
		DDERR_INVALIDCAPS:                  ErrorMessage := 'DDERR_INVALIDCAPS';
		DDERR_INVALIDCLIPLIST:              ErrorMessage := 'DDERR_INVALIDCLIPLIST';
		DDERR_INVALIDMODE:                  ErrorMessage := 'DDERR_INVALIDMODE';
		DDERR_INVALIDOBJECT:                ErrorMessage := 'DDERR_INVALIDOBJECT';
		DDERR_INVALIDPARAMS:                ErrorMessage := 'DDERR_INVALIDPARAMS';
		DDERR_INVALIDPIXELFORMAT:           ErrorMessage := 'DDERR_INVALIDPIXELFORMAT';
		DDERR_INVALIDRECT:                  ErrorMessage := 'DDERR_INVALIDRECT';
		DDERR_LOCKEDSURFACES:               ErrorMessage := 'DDERR_LOCKEDSURFACES';
		DDERR_NO3D:                         ErrorMessage := 'DDERR_NO3D';
		DDERR_NOALPHAHW:                    ErrorMessage := 'DDERR_NOALPHAHW';
		DDERR_NOCLIPLIST:                   ErrorMessage := 'DDERR_NOCLIPLIST';
		DDERR_NOCOLORCONVHW:                ErrorMessage := 'DDERR_NOCOLORCONVHW';
		DDERR_NOCOOPERATIVELEVELSET:        ErrorMessage := 'DDERR_NOCOOPERATIVELEVELSET';
		DDERR_NOCOLORKEY:                   ErrorMessage := 'DDERR_NOCOLORKEY';
		DDERR_NOCOLORKEYHW:                 ErrorMessage := 'DDERR_NOCOLORKEYHW';
		DDERR_NODIRECTDRAWSUPPORT:          ErrorMessage := 'DDERR_NODIRECTDRAWSUPPORT';
		DDERR_NOEXCLUSIVEMODE:              ErrorMessage := 'DDERR_NOEXCLUSIVEMODE';
		DDERR_NOFLIPHW:                     ErrorMessage := 'DDERR_NOFLIPHW';
		DDERR_NOGDI:                        ErrorMessage := 'DDERR_NOGDI';
		DDERR_NOMIRRORHW:                   ErrorMessage := 'DDERR_NOMIRRORHW';
		DDERR_NOTFOUND:                     ErrorMessage := 'DDERR_NOTFOUND';
		DDERR_NOOVERLAYHW:                  ErrorMessage := 'DDERR_NOOVERLAYHW';
		DDERR_NORASTEROPHW:                 ErrorMessage := 'DDERR_NORASTEROPHW';
		DDERR_NOROTATIONHW:                 ErrorMessage := 'DDERR_NOROTATIONHW';
		DDERR_NOSTRETCHHW:                  ErrorMessage := 'DDERR_NOSTRETCHHW';
		DDERR_NOT4BITCOLOR:                 ErrorMessage := 'DDERR_NOT4BITCOLOR';
		DDERR_NOT4BITCOLORINDEX:            ErrorMessage := 'DDERR_NOT4BITCOLORINDEX';
		DDERR_NOT8BITCOLOR:                 ErrorMessage := 'DDERR_NOT8BITCOLOR';
		DDERR_NOTEXTUREHW:                  ErrorMessage := 'DDERR_NOTEXTUREHW';
		DDERR_NOVSYNCHW:                    ErrorMessage := 'DDERR_NOVSYNCHW';
		DDERR_NOZBUFFERHW:                  ErrorMessage := 'DDERR_NOZBUFFERHW';
		DDERR_NOZOVERLAYHW:                 ErrorMessage := 'DDERR_NOZOVERLAYHW';
		DDERR_OUTOFCAPS:                    ErrorMessage := 'DDERR_OUTOFCAPS';
		DDERR_OUTOFMEMORY:                  ErrorMessage := 'DDERR_OUTOFMEMORY';
		DDERR_OUTOFVIDEOMEMORY:             ErrorMessage := 'DDERR_OUTOFVIDEOMEMORY';
		DDERR_OVERLAYCANTCLIP:              ErrorMessage := 'DDERR_OVERLAYCANTCLIP';
		DDERR_OVERLAYCOLORKEYONLYONEACTIVE: ErrorMessage := 'DDERR_OVERLAYCOLORKEYONLYONEACTIVE';
		DDERR_PALETTEBUSY:                  ErrorMessage := 'DDERR_PALETTEBUSY';
		DDERR_COLORKEYNOTSET:               ErrorMessage := 'DDERR_COLORKEYNOTSET';
		DDERR_SURFACEALREADYATTACHED:       ErrorMessage := 'DDERR_SURFACEALREADYATTACHED';
		DDERR_SURFACEALREADYDEPENDENT:      ErrorMessage := 'DDERR_SURFACEALREADYDEPENDENT';
		DDERR_SURFACEBUSY:                  ErrorMessage := 'DDERR_SURFACEBUSY';
		DDERR_CANTLOCKSURFACE:              ErrorMessage := 'DDERR_CANTLOCKSURFACE';
		DDERR_SURFACEISOBSCURED:            ErrorMessage := 'DDERR_SURFACEISOBSCURED';
		DDERR_SURFACELOST:                  ErrorMessage := 'DDERR_SURFACELOST';
		DDERR_SURFACENOTATTACHED:           ErrorMessage := 'DDERR_SURFACENOTATTACHED';
		DDERR_TOOBIGHEIGHT:                 ErrorMessage := 'DDERR_TOOBIGHEIGHT';
		DDERR_TOOBIGSIZE:                   ErrorMessage := 'DDERR_TOOBIGSIZE';
		DDERR_TOOBIGWIDTH:                  ErrorMessage := 'DDERR_TOOBIGWIDTH';
		DDERR_UNSUPPORTED:                  ErrorMessage := 'DDERR_UNSUPPORTED';
		DDERR_UNSUPPORTEDFORMAT:            ErrorMessage := 'DDERR_UNSUPPORTEDFORMAT';
		DDERR_UNSUPPORTEDMASK:              ErrorMessage := 'DDERR_UNSUPPORTEDMASK';
		DDERR_VERTICALBLANKINPROGRESS:      ErrorMessage := 'DDERR_VERTICALBLANKINPROGRESS';
		DDERR_WASSTILLDRAWING:              ErrorMessage := 'DDERR_WASSTILLDRAWING';
		DDERR_XALIGN:                       ErrorMessage := 'DDERR_XALIGN';
		DDERR_INVALIDDIRECTDRAWGUID:        ErrorMessage := 'DDERR_INVALIDDIRECTDRAWGUID';
		DDERR_DIRECTDRAWALREADYCREATED:     ErrorMessage := 'DDERR_DIRECTDRAWALREADYCREATED';
		DDERR_NODIRECTDRAWHW:               ErrorMessage := 'DDERR_NODIRECTDRAWHW';
		DDERR_PRIMARYSURFACEALREADYEXISTS:  ErrorMessage := 'DDERR_PRIMARYSURFACEALREADYEXISTS';
		DDERR_NOEMULATION:                  ErrorMessage := 'DDERR_NOEMULATION';
		DDERR_REGIONTOOSMALL:               ErrorMessage := 'DDERR_REGIONTOOSMALL';
		DDERR_CLIPPERISUSINGHWND:           ErrorMessage := 'DDERR_CLIPPERISUSINGHWND';
		DDERR_NOCLIPPERATTACHED:            ErrorMessage := 'DDERR_NOCLIPPERATTACHED';
		DDERR_NOHWND:                       ErrorMessage := 'DDERR_NOHWND';
		DDERR_HWNDSUBCLASSED:               ErrorMessage := 'DDERR_HWNDSUBCLASSED';
		DDERR_HWNDALREADYSET:               ErrorMessage := 'DDERR_HWNDALREADYSET';
		DDERR_NOPALETTEATTACHED:            ErrorMessage := 'DDERR_NOPALETTEATTACHED';
		DDERR_NOPALETTEHW:                  ErrorMessage := 'DDERR_NOPALETTEHW';
		DDERR_BLTFASTCANTCLIP:              ErrorMessage := 'DDERR_BLTFASTCANTCLIP';
		DDERR_NOBLTHW:                      ErrorMessage := 'DDERR_NOBLTHW';
		DDERR_NODDROPSHW:                   ErrorMessage := 'DDERR_NODDROPSHW';
		DDERR_OVERLAYNOTVISIBLE:            ErrorMessage := 'DDERR_OVERLAYNOTVISIBLE';
		DDERR_NOOVERLAYDEST:                ErrorMessage := 'DDERR_NOOVERLAYDEST';
		DDERR_INVALIDPOSITION:              ErrorMessage := 'DDERR_INVALIDPOSITION';
		DDERR_NOTAOVERLAYSURFACE:           ErrorMessage := 'DDERR_NOTAOVERLAYSURFACE';
		DDERR_EXCLUSIVEMODEALREADYSET:      ErrorMessage := 'DDERR_EXCLUSIVEMODEALREADYSET';
		DDERR_NOTFLIPPABLE:                 ErrorMessage := 'DDERR_NOTFLIPPABLE';
		DDERR_CANTDUPLICATE:                ErrorMessage := 'DDERR_CANTDUPLICATE';
		DDERR_NOTLOCKED:                    ErrorMessage := 'DDERR_NOTLOCKED';
		DDERR_CANTCREATEDC:                 ErrorMessage := 'DDERR_CANTCREATEDC';
		DDERR_NODC:                         ErrorMessage := 'DDERR_NODC';
		DDERR_WRONGMODE:                    ErrorMessage := 'DDERR_WRONGMODE';
		DDERR_IMPLICITLYCREATED:            ErrorMessage := 'DDERR_IMPLICITLYCREATED';
		DDERR_NOTPALETTIZED:                ErrorMessage := 'DDERR_NOTPALETTIZED';
		DDERR_UNSUPPORTEDMODE:              ErrorMessage := 'DDERR_UNSUPPORTEDMODE';
		DDERR_NOMIPMAPHW:                   ErrorMessage := 'DDERR_NOMIPMAPHW';
		DDERR_INVALIDSURFACETYPE:           ErrorMessage := 'DDERR_INVALIDSURFACETYPE';
		DDERR_DCALREADYCREATED:             ErrorMessage := 'DDERR_DCALREADYCREATED';
		DDERR_CANTPAGELOCK:                 ErrorMessage := 'DDERR_CANTPAGELOCK';
		DDERR_CANTPAGEUNLOCK:               ErrorMessage := 'DDERR_CANTPAGEUNLOCK';
		DDERR_NOTPAGELOCKED:                ErrorMessage := 'DDERR_NOTPAGELOCKED';
		DDERR_NOTINITIALIZED:               ErrorMessage := 'DDERR_NOTINITIALIZED';
	else
		ErrorMessage := 'Unknown Error';
	end;

	mmoEvents.Lines.Add('Error - ' + ErrorMessage);
  //ShowMessage(ErrorMessage);
end;

procedure TGraphics.BackSetPixel(Surface: IDirectDrawSurface4; x,
	y: Integer; Color: Byte);
begin

	if (Surface = nil) then
	begin
		exit;
	end;

	backvidbuf[x+y*ddsdsetpixelb.lPitch] := Color;

end;

function TGraphics.BlitNESBackground(HorizontalScroll,
	VerticalScroll: Byte): Boolean;
var
	mybltfx:			TDDBLTFX;
	nilbltfx:			TDDBLTFX;
	ColorKey:			TDDCOLORKEY;
	SourceRect:		TRect;
	DestRect:			TRect;
begin

	mybltfx.dwSize := SizeOf(mybltfx);
	mybltfx.dwFillColor := PPU.BackgroundColor;

	NESScreen.ClearSurface(1);

	ColorKey.dwColorSpaceHighValue := 0;
	ColorKey.dwColorSpaceLowValue := 0;

	g_pDDSNESBack.SetColorKey(DDCKEY_SRCBLT, @ColorKey);

	nilbltfx.dwSize := SizeOf(nilbltfx);

	case SCREEN_WIDTH of
		640:{480}
			begin
				DestRect.Left := 0;
				DestRect.Right := 511;
				DestRect.Top := 0;
				DestRect.Bottom := 480;
			end;
		320:{240}
			begin
				DestRect.Left := 0;
				DestRect.Right := 255;
				DestRect.Top := 0;
				DestRect.Bottom := 240;
			end;
		end;

	SourceRect.Left := 0;
	SourceRect.Right := 255;
	SourceRect.Top := 0;
	SourceRect.Bottom := 240;

	SourceRect.Left := SourceRect.Left + HorizontalScroll;
	SourceRect.Right := SourceRect.Right + HorizontalScroll;
	SourceRect.Top := SourceRect.Top + VerticalScroll;
	SourceRect.Bottom := SourceRect.Bottom + VerticalScroll;

	g_pDDSBack.Blt(@DestRect, g_pDDSNESBack, @SourceRect, (DDBLT_WAIT or DDBLT_KEYSRC),@nilbltfx);
	Result := True;

end;

procedure TGraphics.CheckSurfaces;
begin
	if ((g_pDDSPrimary.IsLost) = DDERR_SURFACELOST) then
	begin
		g_pDDSPrimary._Restore;
	end;
	if ((g_pDDSBack.IsLost) = DDERR_SURFACELOST) then
	begin
		g_pDDSBack._Restore;
	end;
	if ((g_pDDSNESBack.IsLost) = DDERR_SURFACELOST) then
	begin
		g_pDDSNESBack._Restore;
	end;
end;

function TGraphics.ClearSurface(Mode: Integer): Boolean;
var
	BltFx: 				TDDBLTFX;
	rc:						TRect;
	Surface:			IDirectDrawSurface4;
begin
	Result := True;
	try
		rc.Top := 0;
		rc.Left := 0;

		if Mode = 1 then
		begin
			rc.Right := SCREEN_WIDTH;
			rc.Bottom := SCREEN_HEIGHT;
			Surface := g_pDDSBack;
		end else
		begin
			rc.Right := 512;
			rc.Bottom := 480;
			Surface := g_pDDSNESBack;
		end;

		ZeroMemory ( @BltFx, sizeof(BltFx));
		BltFx.dwSize := sizeof(BltFx);
		BltFx.dwFillColor := PPU.BackgroundColor;	// Black
		Surface.Blt(@rc, nil, @rc, DDBLT_COLORFILL + DDBLT_WAIT, @BltFx);
	except
		Result := False;
	end;

end;

constructor TGraphics.Create;
begin
	{}
end;

function TGraphics.CreateSurfaces(bFullScreen: Boolean): Boolean;
var
	HR:					HRESULT;
	ddsd:      	TDDSURFACEDESC2;
	ddscaps:		TDDSCAPS2;
begin
	FullScreen := bFullScreen;

	if (FullScreen) then
	begin
		Screen.Cursor := crNone;
		if frmSettings.rb640.Checked then
		begin
			SCREEN_WIDTH := 640;
			SCREEN_HEIGHT := 480;
		end;
		if frmSettings.rb320.Checked then
		begin
			SCREEN_WIDTH := 320;
			SCREEN_HEIGHT := 240;
		end;

		//Full screen
		HR := g_pDD.SetCooperativeLevel(WinHandle,DDSCL_EXCLUSIVE or DDSCL_FULLSCREEN or DDSCL_ALLOWREBOOT);
		if (HR <> DD_OK) then
		begin
    	frmCMNES.DXError(HR);
			Result := False;
			exit;
		end;

		HR := g_pDD.SetDisplayMode(SCREEN_WIDTH,SCREEN_HEIGHT,8,0,0);
		if (HR <> DD_OK) then
		begin
    	frmCMNES.DXError(HR);
			Result := False;
			exit;
		end;
	end else
	begin
		//Windowed
		SCREEN_WIDTH := 320;
		SCREEN_HEIGHT := 240;
		HR := g_pDD.SetCooperativeLevel(WinHandle,DDSCL_NORMAL);
		if (HR <> DD_OK) then
		begin
    	frmCMNES.DXError(HR);
			Result := False;
			exit;
		end;
	end;

	ZeroMemory(@ddsd,SizeOf(ddsd));
	ddsd.dwSize := SizeOf(ddsd);

	if (FullScreen) then
	begin
		// Get the dimensions of the viewport and screen bounds
		// Store the rectangle which contains the renderer
		SetRect(g_rcViewport, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT );
		CopyMemory(@g_rcScreen, @g_rcViewport, SizeOf(TRect));

		// Create the primary surface with 1 back buffer
		ZeroMemory(@ddsd,SizeOf(ddsd));
		ddsd.dwSize := SizeOf(ddsd);
		ddsd.dwFlags := DDSD_CAPS or
									 DDSD_BACKBUFFERCOUNT;
		ddsd.ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE or
										DDSCAPS_FLIP or
										DDSCAPS_COMPLEX;
		ddsd.dwBackBufferCount := 1;
		HR := g_pDD.CreateSurface(ddsd, g_pDDSPrimary, nil);
		if (HR <> DD_OK) then
		begin
    	frmCMNES.DXError(HR);
		end;

		ddscaps.dwCaps := DDSCAPS_BACKBUFFER;
		HR := g_pDDSPrimary.GetAttachedSurface(ddscaps, g_pDDSBack);
		if (HR <> DD_OK) then
		begin
    	frmCMNES.DXError(HR);
		end;
			// now create the surface to hold the pattern table
		ZeroMemory(@ddsd, sizeof(ddsd));
		ddsd.dwSize := SizeOf(ddsd);
		ddsd.dwFlags := DDSD_CAPS or DDSD_WIDTH or DDSD_HEIGHT;
		ddsd.dwWidth := 512;		// two screens wide
		ddsd.dwHeight := 480;	// two screens tall
		if (g_pDD.CreateSurface(ddsd, g_pDDSNesBack ,nil) <> DD_OK) then
		begin
			ShowMessage('CreateSurface 3 failed');
		end;
	end else
	begin
		//Create primary surface
		ddsd.dwFlags := DDSD_CAPS;
		ddsd.ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE;
		HR := g_pDD.CreateSurface(ddsd,g_pDDSPrimary,nil);
		if (HR <> DD_OK) then
		begin
    	frmCMNES.DXError(HR);
			Result := False;
			exit;
		end;
		//Create back surface
		ddsd.dwFlags := DDSD_WIDTH or DDSD_HEIGHT or DDSD_CAPS;
		ddsd.dwWidth := SCREEN_WIDTH;
		ddsd.dwHeight := SCREEN_HEIGHT;
		ddsd.ddsCaps.dwCaps := DDSCAPS_OFFSCREENPLAIN;
		HR := g_pDD.CreateSurface(ddsd,g_pDDSBack,nil);
		if (HR <> DD_OK) then
		begin
    	frmCMNES.DXError(HR);
			Result := False;
			exit;
		end;
		//Create NES back surface
		ddsd.dwFlags := DDSD_WIDTH or DDSD_HEIGHT or DDSD_CAPS;
		ddsd.dwWidth := 512;
		ddsd.dwHeight := 480;
		ddsd.ddsCaps.dwCaps := DDSCAPS_OFFSCREENPLAIN;
		HR := g_pDD.CreateSurface(ddsd,g_pDDSNESBack,nil);
		if (HR <> DD_OK) then
		begin
    	frmCMNES.DXError(HR);
			Result := False;
			exit;
		end;

	end;

	//Clipper
	if (not FullScreen) then
	begin
		HR := g_pDD.CreateClipper(0,g_pClipper,nil);
		if (HR <> DD_OK) then
		begin
    	frmCMNES.DXError(HR);
			Result := False;
			exit;
		end;
		HR := g_pClipper.SetHWnd(0,WinHandle);
		if (HR <> DD_OK) then
		begin
    	frmCMNES.DXError(HR);
			Result := False;
			exit;
		end;
		HR := g_pDDSPrimary.SetClipper(g_pClipper);
		if (HR <> DD_OK) then
		begin
    	frmCMNES.DXError(HR);
			Result := False;
			exit;
		end;
	end;
  
	Result := True;

end;

destructor TGraphics.Destroy;
begin
  if g_pClipper <> nil then
  	g_pClipper._Release;

	if g_pDDSNESBack <> nil then
		g_pDDSNESBack._Release;

  if g_pDDSBack <> nil then
		g_pDDSBack._Release;

	if g_pDDSPrimary <> nil then
		g_pDDSPrimary._Release;

  if g_pDD <> nil then
  	g_pDD._Release;

	if pDD <> nil then
  	pDD._Release;

end;

procedure TGraphics.DrawFrame;
begin
	Flip;
	CheckSurfaces;
end;

procedure TGraphics.Flip;
var
	HR:					HRESULT;
	rcSrc:			TRect;
	rcDest:			TRect;
	p:					TPoint;
	myBltFx:		TDDBLTFX;
	myHDC:			HDC;
begin
	if not (FullScreen) then
	begin
		p.x := 0;
		p.y := 0;
		myBltFx.dwSize := SizeOf(TDDBLTFX);
		ClientToScreen(WinHandle,p);
		GetClientRect(WinHandle,rcDest);
		OffsetRect(rcDest,p.x,p.y);
		SetRect(rcSrc,0,0,SCREEN_WIDTH,SCREEN_HEIGHT);
		HR := g_pDDSPrimary.Blt(@rcDest,g_pDDSBack,@rcSrc,0,@myBltFx);
		if (HR <> DD_OK) then
		begin
			frmCMNES.DXError(HR);
		end;
	end else
	begin
		HR := g_pDDSPrimary.Flip(nil,DDFLIP_WAIT);
		if (HR <> DD_OK) then
		begin
			frmCMNES.DXError(HR);
		end;
	end;

	{frames per second}
	if frmCMNES.mnuFramesPerSecond.Checked then
	begin
		myHDC := GetDC(NESScreen.WinHandle);
		frmCMNES.Font.Size := 8;
		TextOut(myHDC,5,frmCMNES.Height - 68,PChar(strFrameCount),Length(strFrameCount));
	end;

end;

function TGraphics.Initialise(Handle: HWND): Boolean;
var
	HR:		HRESULT;
begin
//	Initialise DirectDraw

	//Result := True;

	try
		WinHandle := Handle; {store handle}
		if (WinHandle = 0) then
		begin
			ShowMessage('Handle error');
		end;

		pDD := nil;
		HR := DirectDrawCreate(nil, pDD, nil);
		if (HR <> DD_OK) then
		begin
			ShowMessage('DirectDrawCreate failed');
			Result := False;
		end else
		begin
			Result := True;
		end;

		// Fetch DirectDraw4 interface
		HR := pDD.QueryInterface(IID_IDirectDraw4, g_pDD);
		if (HR <> DD_OK) then
		begin
			ShowMessage('QueryInterface failed');
		end;

		if frmCMNES.mnuWindowed.Checked then
		begin
			CreateSurfaces(False);
		end else
		begin
			CreateSurfaces(True);
		end;

		pDD := nil;

	except
		mmoEvents.Lines.Add('DirectX - Initialization Failed');
		Result := False;
		exit;
	end;

	mmoEvents.Lines.Add('DirectX - Initialization Successful');

end;

procedure TGraphics.InitPalette;
var
	palette: 				Array [0..256] of PALETTEENTRY;
	iX:							Integer;
	HR:							HRESULT;
type
	TPalType = array[0..63, 1..3] of byte;
const
	Pal: TPalType = {Alex Krasivsky's Palette}
		(
		($7f,$7f,$7f), ($00,$00,$FF), ($00,$00,$BF), ($47,$2b,$bf),
		($97,$00,$87), ($ab,$00,$23), ($ab,$13,$00), ($8b,$17,$00),
		($53,$30,$00), ($00,$78,$00), ($00,$6b,$00), ($00,$5b,$00),
		($00,$43,$58), ($00,$00,$00), ($00,$00,$00), ($00,$00,$00),

		($bf,$bf,$bf), ($00,$78,$F8), ($00,$58,$F8), ($6B,$47,$ff),
		($db,$00,$cd), ($e7,$00,$5b), ($F8,$38,$00), ($e7,$5f,$13),
		($af,$7f,$00), ($00,$B8,$00), ($00,$ab,$00), ($00,$ab,$47),
		($00,$8b,$8b), ($00,$00,$00), ($00,$00,$00), ($00,$00,$00),

		($F8,$F8,$F8), ($3f,$bf,$FF), ($6b,$88,$FF), ($98,$78,$F8),
		($F8,$78,$f8), ($F8,$58,$98), ($F8,$78,$58), ($FF,$a3,$47),
		($F8,$b8,$00), ($b8,$f8,$18), ($5b,$Db,$57), ($58,$f8,$98),
		($00,$eb,$db), ($78,$78,$78), ($00,$00,$00), ($00,$00,$00),

		($FF,$FF,$FF), ($a7,$e7,$FF), ($B8,$B8,$F8), ($d8,$B8,$F8),
		($F8,$B8,$f8), ($Fb,$a7,$C3), ($F0,$d0,$B0), ($FF,$e3,$Ab),
		($Fb,$Db,$7b), ($d8,$f8,$78), ($b8,$f8,$B8), ($b8,$F8,$d8),
		($00,$ff,$FF), ($f8,$D8,$f8), ($00,$00,$00), ($00,$00,$00));

begin

	ZeroMemory(@palette, 256 * sizeof(PALETTEENTRY));

	// Load NES Palette;

	for iX := 0 to 63 do
	begin
		palette[iX].peRed := PAL[iX,1];
		palette[iX].peGreen := PAL[iX,2];
		palette[iX].peBlue := PAL[iX,3];
		palette[iX].peFlags := PC_NOCOLLAPSE;
	end;

	if ((g_pDD.CreatePalette(DDPCAPS_8BIT or DDPCAPS_INITIALIZE or DDPCAPS_ALLOW256,
															@palette,
															lpddpal,
															nil)) <> DD_OK) then
	begin
		ShowMessage('Error creating palette!');
	end	else
	begin
		HR := (g_pDDSPrimary.SetPalette(lpddpal));
		if (HR <> DD_OK) then
		begin
			if (HR = 2289434769) then
			begin
				ShowMessage('You need to switch to 256 colour mode!');
				mmoEvents.Lines.Add('DirectX - Not 256 Colour Mode');
			end else
			begin
	    	frmCMNES.DXError(HR);
			end;
		end;
	end;

	{update palette}

	lpddpal._Release;

end;

procedure TGraphics.LockBackground(Surface: IDirectDrawSurface4);
begin
	ZeroMemory(@ddsdsetpixel, SizeOf(ddsdsetpixel));
	ddsdsetpixel.dwSize := SizeOf(ddsdsetpixel);
	Surface.Lock(nil, ddsdsetpixel, DDLOCK_SURFACEMEMORYPTR or DDLOCK_WAIT, 0);

	vidbuf := ddsdsetpixel.lpSurface;
end;

procedure TGraphics.LockNESBackground(Surface: IDirectDrawSurface4);
begin
	ZeroMemory(@ddsdsetpixelb, SizeOf(ddsdsetpixelb));
	ddsdsetpixelb.dwSize := SizeOf(ddsdsetpixelb);
	Surface.Lock(nil, ddsdsetpixelb, DDLOCK_SURFACEMEMORYPTR or DDLOCK_WAIT, 0);
	backvidbuf := ddsdsetpixelb.lpSurface;
end;

procedure TGraphics.SetPixel(Surface: IDirectDrawSurface4; x, y: Integer;
	Color: Byte);
begin

	if (Surface = nil) then
	begin
		exit;
	end;

	//y := y + 1;

	case SCREEN_WIDTH of
		640:{480}
			begin
				VidBuf[x*2+y*2 *ddsdsetpixel.lPitch] := Color;
				VidBuf[x*2+y*2 *ddsdsetpixel.lPitch +1] := Color;
				vidbuf[x*2+y*2*ddsdsetpixel.lPitch + ddsdsetpixel.lPitch] := Color;
				vidbuf[x*2+y*2*ddsdsetpixel.lPitch + ddsdsetpixel.lPitch + 1] := Color;
			end;
		320:{240}
			begin
				vidbuf[x+y*ddsdsetpixel.lPitch] := Color;
			end;
		end;

end;

procedure TGraphics.SetPixelBehindBackground(x, y: Integer; Color, BackgroundColor: Byte);
begin

	case SCREEN_WIDTH of
		640:{480}
			begin
				if (VidBuf[x*2+y*2 * ddsdsetpixel.lPitch] = BackgroundColor) then
				begin
					VidBuf[x*2+y*2 *ddsdsetpixel.lPitch] := Color;
					VidBuf[x*2+y*2 *ddsdsetpixel.lPitch +1] := Color;
					vidbuf[x*2+y*2*ddsdsetpixel.lPitch + ddsdsetpixel.lPitch] := Color;
					vidbuf[x*2+y*2*ddsdsetpixel.lPitch + ddsdsetpixel.lPitch + 1] := Color;
				end;
			end;
		320:{240}
			begin
				if (VidBuf[x+y * ddsdsetpixel.lPitch] = BackgroundColor) then
					VidBuf[x+y *ddsdsetpixel.lPitch] := Color;
			end;
		end;

end;

procedure TGraphics.UnlockSurface(Surface: IDirectDrawSurface4);
begin
	{unlock surface}
	Surface.Unlock(nil);
end;


//***************************//
//	TInput                   //
//***************************//
constructor TInput.Create;
var
	HR:								HRESULT;
	TempDevice: 			IDirectInputDevice2;
	iCount: 					Integer;
	DevCaps: 					TDIDevCaps;
	RangeProp: 				TDIPropRange;
	DataProp: 				TDIPropDWord;
	DIPropAutoCenter: TDIPropDWord;
begin
	{create the DirectInput object}
	HR :=	DirectInputCreateEx(hInstance, DIRECTINPUT_VERSION,IID_IDirectInput7, pDI, nil);

	if HR <> DI_OK then
		frmCMNES.DXError(HR);

	{create a list of devices}
	JoyList := TStringList.Create;

	{enumerate attached devices}
	HR := pDI.EnumDevices(DIDEVTYPE_JOYSTICK, EnumInputDevs, JoyList, DIEDFL_ALLDEVICES);

	if HR <> DI_OK then
		frmCMNES.DXError(HR);

  if JoyList.Count > 0 then
	begin
		HR := pDI.CreateDeviceEx(PGUID(JoyList.Objects[0])^,IID_IDirectInputDevice2, TempDevice, nil);

    if HR <> DI_OK then
      frmCMNES.DXError(HR);

    {we need the new version of the input device interface, so retrieve it}
    try
      TempDevice.QueryInterface(IDirectInputDevice2, pDIJoystick);
    finally
      TempDevice := nil
    end;

    {retrieve the capabilities of the device}
    DevCaps.dwSize := SizeOf(DevCaps);
    pDIJoystick.GetCapabilities(DevCaps);

    {initialize our axis tracking variable}
    NumAxes := 0;

		{now we need to begin creating the data format structure and setting the
		data format, so we must initialize the overall data format structure}
		with FDataFormat do
		begin
			dwFlags := DIDF_ABSAXIS;
      dwDataSize := SizeOf(TDIJoyState2);
      dwSize := SizeOf(FDataFormat);
      dwObjSize := SizeOf(TDIObjectDataFormat);
      dwNumObjs := 0;
      rgodf := @FDataFormatObjects;
    end;

    {create and set a data format specifically made for this device}
    SetDataFormat;

		{we must now set the range for each axis, so initialize a range
     property structure}
    RangeProp.diph.dwSize := SizeOf(TDIPropRange);
		RangeProp.diph.dwHeaderSize := SizeOf(TDIPropHeader);
		RangeProp.diph.dwHow := DIPH_BYOFFSET;
		RangeProp.lMin := -1000;
    RangeProp.lMax := 1000;

		{we must also set deadzones and saturations for each axis, so
     initialize a DWord property structure}
    DataProp.diph.dwSize := SizeOf(TDIPropDWord);
    DataProp.diph.dwHeaderSize := SizeOf(TDIPropHeader);
    DataProp.diph.dwHow := DIPH_BYOFFSET;

    {for each axis...}
    for iCount := 0 to NumAxes-1 do
    begin
      {set its range...}
      RangeProp.diph.dwObj := FAxisList[iCount];
			pDIJoystick.SetProperty(DIPROP_RANGE, RangeProp.diph);

      {its deadzone... (this value is in 100ths of a percent}
			DataProp.dwData := 1000;
			DataProp.diph.dwObj := FAxisList[iCount];
			pDIJoystick.SetProperty(DIPROP_DEADZONE, DataProp.diph);

      {and its saturation (this value is also in 100ths of a percent}
			DataProp.dwData := 9000;
      pDIJoystick.SetProperty(DIPROP_SATURATION, DataProp.diph);
    end;

		{set the cooperative level.}
		pDIJoystick.SetCooperativeLevel(hInstance, DISCL_EXCLUSIVE or DISCL_FOREGROUND);

		{turn off auto center}
		DIPropAutoCenter.diph.dwSize := SizeOf(TDIPROPDWORD);
		DIPropAutoCenter.diph.dwHeaderSize := SizeOf(TDIPROPHEADER);
		DIPropAutoCenter.diph.dwObj := 0;
		DIPropAutoCenter.diph.dwHow := DIPH_DEVICE;
		DIPropAutoCenter.dwData := DIPROPAUTOCENTER_OFF;

		HR := pDIJoystick.SetProperty(DIPROP_AUTOCENTER,DIPropAutoCenter.diph);

		if HR <> DI_OK then
      frmCMNES.DXError(HR);

		{create ffb effects}
		(*CreateEffects;

		HR := pDIJoystick.CreateEffect(GUID_ConstantForce,@Effect1,pConstantForce,nil);

		if HR <> DI_OK then
			frmCMNES.DXError(HR);*)

		{finally, acquire the device so we can start retrieving information}
		pDIJoystick.Acquire;

		pConstantForce.Start(1,0);

	end;

	{create keyboard}
	HR := pDI.CreateDevice(GUID_SysKeyboard,pDIKeyboard,nil);

	if HR <> DI_OK then
		frmCMNES.DXError(HR);

	HR := pDIKeyboard.SetCooperativeLevel(hInstance, DISCL_BACKGROUND or DISCL_NONEXCLUSIVE);

	if HR <> DI_OK then
		frmCMNES.DXError(HR);

	{set data format}
	pDIKeyboard.SetDataFormat(c_dfDIKeyboard);

	if HR <> DI_OK then
		frmCMNES.DXError(HR);

	{aquire keyboard}
	pDIKeyboard.Acquire;

	if HR <> DI_OK then
		frmCMNES.DXError(HR);

	InputType := 1;
end;

function TInput.CreateEffects: Boolean;
begin
	Result := True;

  try
    ZeroMemory(@Effect1,SizeOf(Effect1));

		dwAxes[0] := DIJOFS_X;
		dwAxes[1] := DIJOFS_Y;

		direction[0] := 90 * DI_DEGREES;
		direction[1] := 90 * DI_DEGREES;

		diCF.lMagnitude := DI_FFNOMINALMAX;

		Effect1.dwSize := SizeOf(TDIEFFECT);
		Effect1.dwFlags := DIEFF_POLAR or DIEFF_OBJECTOFFSETS;
		Effect1.dwDuration := 10 * DI_SECONDS;
		Effect1.dwSamplePeriod := 0;
		Effect1.dwGain := DI_FFNOMINALMAX;
		Effect1.dwTriggerButton := DIEB_NOTRIGGER;
		Effect1.dwTriggerRepeatInterval := 0;
		Effect1.cAxes := 2;
		Effect1.rgdwAxes := @dwAxes;
		Effect1.rglDirection := @direction;
		Effect1.lpEnvelope := nil;
		Effect1.cbTypeSpecificParams := SizeOf(TDICONSTANTFORCE);
		Effect1.lpvTypeSpecificParams := @diCF;
	except
		Result := False;
	end;
end;

destructor TInput.Destroy;
begin

	if (pDIKeyboard <> nil) then
	begin
		pDIKeyboard._Release;
		//pDIKeyboard := nil;
  end;

	if (pConstantForce <> nil) then
		pConstantForce._Release;

	if (pDIJoystick <> nil) then
	begin
		pDIJoystick._Release;
	end;

	if (pDI <> nil) then
	begin
  	pDI._Release;
    //pDI := nil;
  end;

end;

function TInput.ReadPad(Pad, Read: Byte): Byte;
var
	Button:			Integer;
begin

	{device read performed earlier through polling}

	case InputType of
		1:	{standard pad}
			begin
				if Pad = 1 then {pad 1}
				begin
					Case Read of
						1:
							begin
								Button := Keystate[DIK_N] and $80;
								Button := Button  or JoyState.rgbButtons[2] and $80;
							end;
						2:
							begin
								Button := Keystate[DIK_M] and $80;
								Button := Button or JoyState.rgbButtons[3] and $80;
							end;
						3:
							begin
								Button := Keystate[DIK_F1] and $80;
								Button := Button or JoyState.rgbButtons[0] and $80;
							end;
						4:
							begin
								Button := Keystate[DIK_F2] and $80;
								Button := Button or JoyState.rgbButtons[1] and $80;
							end;
						5:
							begin
								Button := Keystate[DIK_UP] and $80;
								if JoyState.lY < -250 then
									Button := $80;
							end;
						6:
							begin
								Button :=	Keystate[DIK_DOWN] and $80;
								if JoyState.lY > 250 then
									Button := $80;
							end;
						7:
							begin
								Button := Keystate[DIK_LEFT] and $80;
								if JoyState.lX < -250 then
									Button := $80;
							end;
						8:
							begin
								Button := Keystate[DIK_RIGHT] and $80;
								if JoyState.lX > 250 then
									Button := $80;
							end;
						else
							Button := 0;
						end;
					if Button <> 0 then
					begin
						Result := $41;
					end else
					begin
						Result := $40;
					end;
				end else {pad 2}               
				begin
					Case Read of
						1:
							Button := Keystate[DIK_C] and $80;
						2:
							Button := Keystate[DIK_V] and $80;
						3:
							Button := Keystate[DIK_F3] and $80;
						4:
							Button := Keystate[DIK_F4] and $80;
						5:
							Button := Keystate[DIK_Q] and $80;
						6:
							Button :=	Keystate[DIK_A] and $80;
						7:
							Button := Keystate[DIK_Z] and $80;
						8:
							Button := Keystate[DIK_X] and $80;
						else
							Button := 0;
						end;
					if Button <> 0 then
					begin
						Result := $41;
					end else
					begin
						Result := $40;
					end;
				end;
			end;
		else
			Result := $00;
		end;

end;

//***************************//
//	TGamePad                 //
//***************************//
constructor TGamePad.Create;
begin
	ResetValue := 0;		// make sure it starts out clear
	Strobe := 0;
	ReadNum := 0;
	PadNumber := 0;
end;

function TGamePad.ReadByte: Byte;
begin
	if ResetValue = 0 then
	begin
		ShowMessage('Attempt to read uninitialised controller');
		Result := $00;
	end else
	begin
		Inc(ReadNum);
		if (ReadNum <= 8) then
			Result := Input.ReadPad(PadNumber,ReadNum)
		else if ( (ReadNum = 20) and (PadNumber = 1) ) then	// for the pad 0 signature
			Result := $01
		else if ( (ReadNum = 19) and (PadNumber = 2) ) then	// for the pad 1 signature
			Result := $01
		else
			Result := $00;
	end;

end;

procedure TGamePad.SetPad(Port: Integer);
begin
	PadNumber := Port;
end;

procedure TGamePad.WriteByte(Address: Word;Value: Byte);
begin
	if ( (Address = $4016) or (Address = $4017) ) then	// ignore other writes
	begin
		if ((Strobe = 1) and (Value = 0)) then	// this transition resets the gamepad
		begin
			GamePad1.ResetValue := 1;
			GamePad2.ResetValue := 1;
			GamePad1.Strobe := 0;
			GamePad2.Strobe := 0;
			GamePad1.ReadNum := 0;
			GamePad2.ReadNum := 0;
		end	else
		begin
			if (value = 1) then
			begin
				GamePad1.Strobe := 1;
				GamePad2.Strobe := 1;
			end;
		end;
	end;

end;

//***************************//
//	TfrmCMNES                //
//***************************//
procedure TfrmCMNES.FormCreate(Sender: TObject);
begin
	bBackgroundEnabled := True;
	bSpritesEnabled := True;
	{create event log}
	mmoEvents := TMemo.Create(frmCMNES);
	mmoEvents.Parent := frmCMNES;
	mmoEvents.Visible := False;
	{show title}
	imgTitle.Visible := True;
	{default cycles}
	CycleCount := 0;
	TotalCycles := 0;
	{get settings}
	GetRegistrySettings;
  hWindow := frmCMNES.Handle;
end;

procedure TfrmCMNES.mnuDrawBackgroundClick(Sender: TObject);
begin
	mnuDrawBackground.Checked := not mnuDrawBackground.Checked;
	bBackgroundEnabled := mnuDrawBackground.Checked;
end;

procedure TfrmCMNES.mnuDrawSpritesClick(Sender: TObject);
begin
	mnuDrawSprites.Checked := not mnuDrawSprites.Checked;
	bSpritesEnabled := mnuDrawSprites.Checked;
end;

procedure TfrmCMNES.mnuClearScreenClick(Sender: TObject);
begin
	NESScreen.ClearSurface(1);
	NESScreen.DrawFrame;
end;

procedure TfrmCMNES.mnuCPUResetClick(Sender: TObject);
begin
	CPU.Reset;
	NESScreen.ClearSurface(1);
	NESScreen.DrawFrame;
end;

procedure TfrmCMNES.mnuWindowedClick(Sender: TObject);
begin
	mnuWindowed.Checked := not mnuWindowed.Checked;
	NESScreen.FullScreen := mnuWindowed.Checked;
end;

procedure TfrmCMNES.mnuDisplayPatternTableClick(Sender: TObject);
begin
	NESScreen.ClearSurface(1);
	PPU.DisplayPatternTable;
	NESScreen.Flip;
end;

procedure TfrmCMNES.mnuDisplayPaletteClick(Sender: TObject);
begin
	NESScreen.ClearSurface(1);
	PPU.DisplayPalette;
	NESScreen.Flip;
end;

procedure TfrmCMNES.mnuEventLogClick(Sender: TObject);
begin
	with TfrmEventLog.Create(Application) do
	begin
		try
			mmoEventLog.Lines := mmoEvents.Lines;
			ShowModal;
		finally
			free;
		end;
	end;
end;

procedure TfrmCMNES.mnuDebugClick(Sender: TObject);
begin
	frmDebug.Visible := True;
	frmDebug.Show;
end;

procedure TfrmCMNES.UpdateDebug;
begin
	with frmDebug do
	begin
		if cbUpdateDisplay.Checked then
		begin
			lblRegAccVal.Caption := IntToHex(ufrmCMNES.CPU.RegAcc,2);
			lblRegXVal.Caption := IntToHex(CPU.RegX,2);
			lblRegYVal.Caption := IntToHex(CPU.RegY,2);
			lblPCVal.Caption := IntToHex(CPU.PC,4);
			lblIRVal.Caption := IntToHex(CPU.IR,2);
			lblSRVal.Caption := IntToHex(CPU.SR,2);
			lblSPVal.Caption := IntToHex(CPU.SP,4);
			lblCyclesVal.Caption := IntToStr(CycleCount);

			//	SR
			lblCarryVal.Caption := IntToStr(CPU.BitSet(SR_CARRY_FLAG));
			lblZeroVal.Caption := IntToStr(CPU.BitSet(SR_ZERO_FLAG));
			lblInterruptVal.Caption := IntToStr(CPU.BitSet(SR_INTR_FLAG));
			lblDecimalVal.Caption := IntToStr(CPU.BitSet(SR_DECIMAL_FLAG));
			lblBreakVal.Caption := IntToStr(CPU.BitSet(SR_BREAK_FLAG));
			lblReservedVal.Caption := IntToStr(CPU.BitSet(SR_RSV_FLAG));
			lblOverflowVal.Caption := IntToStr(CPU.BitSet(SR_OVERFLOW_FLAG));
			lblNegativeVal.Caption := IntToStr(CPU.BitSet(SR_NEG_FLAG));
			lblTotalCycles.Caption := IntToStr(TotalCycles);

			lblPPU1.Caption := IntToHex(PPU.PPURegs[0],2);
			lblPPU2.Caption := IntToHex(PPU.PPURegs[1],2);
			lblPPU3.Caption := IntToHex(PPU.PPURegs[2],2);
			lblPPU4.Caption := IntToHex(PPU.PPURegs[3],2);
			lblPPU5.Caption := IntToHex(PPU.PPURegs[4],2);
			lblPPU6.Caption := IntToHex(PPU.PPURegs[5],2);
			lblPPU7.Caption := IntToHex(PPU.PPURegs[6],2);
			lblPPU8.Caption := IntToHex(PPU.PPURegs[7],2);

			lblVRAMAddressVal.Caption := IntToHex(PPU.VRAMAddrReg,4);

			lblScanLineVal.Caption := IntToStr(ScanLine);

			{rom glance}

			lblAddress1.Caption := '$' + IntToHex(CPU.PC,4);
			lblAddress2.Caption := '$' + IntToHex(CPU.PC+1,4);
			lblAddress3.Caption := '$' + IntToHex(CPU.PC+2,4);
			lblAddress4.Caption := '$' + IntToHex(CPU.PC+3,4);
			lblAddress5.Caption := '$' + IntToHex(CPU.PC+4,4);
			lblAddress6.Caption := '$' + IntToHex(CPU.PC+5,4);
			lblAddress7.Caption := '$' + IntToHex(CPU.PC+6,4);
			lblAddress8.Caption := '$' + IntToHex(CPU.PC+7,4);

			lblVal1.Caption := IntToHex(MEMBUS.ReadByte(CPU.PC),2);
			lblVal2.Caption := IntToHex(MEMBUS.ReadByte(CPU.PC+1),2);
			lblVal3.Caption := IntToHex(MEMBUS.ReadByte(CPU.PC+2),2);
			lblVal4.Caption := IntToHex(MEMBUS.ReadByte(CPU.PC+3),2);
			lblVal5.Caption := IntToHex(MEMBUS.ReadByte(CPU.PC+4),2);
			lblVal6.Caption := IntToHex(MEMBUS.ReadByte(CPU.PC+5),2);
			lblVal7.Caption := IntToHex(MEMBUS.ReadByte(CPU.PC+6),2);
			lblVal8.Caption := IntToHex(MEMBUS.ReadByte(CPU.PC+7),2);

			lblHighVal.Caption := IntToStr(Mapper0.M1HighBank);
			lblLowVal.Caption := IntToStr(Mapper0.M1LowBank);
			lblDecodedLow.Caption := IntToHex(Mapper0.DecodedLow,6);
			lblDecodedHigh.Caption := IntToHex(Mapper0.DecodedHigh,6);

		end;

		Application.ProcessMessages;

		{break ?}
		if (edtBreak.Text = IntToHex(CPU.PC,4)) then
		begin
			if not Visible then
			begin
				Visible := True;
			end;
				mnuHalt.Click;
				ShowMessage('Break');
		end;
	end;

end;

procedure TfrmCMNES.ExecuteStep;
begin
	frmCMNES.Caption := 'CMNES - ' + strROMFileName + ' (step)';
	mmoEvents.Lines.Add('CPU - Step');

	Application.ProcessMessages;
	CPU.ExecuteInstruction(CycleCount);
	TotalCycles := TotalCycles + CycleCount;
	if TotalCycles > 1000000 then
	begin
		TotalCycles := 0;
	end;

	if PPU.SpriteRam[0] = ScanLine then
		PPU.PPURegs[2] := PPU.PPURegs[2] or $40; {hit}

	if (CycleCount >= HBLANK) then
	begin
		CycleCount := CycleCount - HBLANK;
		Inc(ScanLine);
		if (ScanLine = SCANLINES) then		// we did the last one already
		begin
			PPU.StartVBlank();	// a vblank should be performed
			if (PPU.ReadByte($2000) and $80) > 0 then
			begin
				CPU.VBlankInterrupt(CycleCount);	// execute an interrupt
			end;
		end;
		if (ScanLine = SCANLINES + VPERIOD/HBLANK) then
		begin
			ScanLine := 0;			// go back to top of screen.
			PPU.EndVBlank();	// end of vblank
		end;
	end;
	if frmDebug.Visible then
	begin
		UpdateDebug;
	end;

end;

procedure TfrmCMNES.ClickAddress(Sender: TLabel);
begin
	with frmDebug do
	begin
		if Length(Sender.Caption) > 4 then
		begin
			edtBreak.Text := Copy(Sender.Caption,2,4);
		end else
		begin
			if Length(Sender.Caption) = 2 then
			begin
				edtBreak.Text := '00' + Sender.Caption;
			end;
		end;
	end;
end;

procedure TfrmCMNES.mnuCaptureScreenClick(Sender: TObject);
var
	DeskDC: 			HDC;
	bmpCapture:		TBitmap;
	SearchRec: 		TSearchRec;
	iNextFileNo:	Integer;
	iResult:			Integer;
	strChop:			String;
	iActualX:			Integer;
	iActualY:			Integer;
begin
	{get first file number}
	iResult := FindFirst(strScreenCapturePath + '\screen*.bmp', faAnyFile, SearchRec);
	if iResult <> 0 then
	begin
		iNextFileNo := 0;
	end else
	begin
		while iResult = 0 do
		begin
			 iResult := FindNext(SearchRec);
		end;
		{get next file number}
		strChop := Copy(SearchRec.Name,7,3);
		strChop := Copy(strChop,1,Pos('.',strChop)-1);
		iNextFileNo := StrToInt(strChop) + 1;
	end;

	{pause for 100ms}
	Sleep(100);
	{capture screen}
	DeskDC := GetDC(GetDesktopWindow);
	iActualX := Round(79.8 / 100 * frmCMNES.Width) - 3;
	if NESScreen.FullScreen then
	begin
		iActualY := frmCMNES.Height;
	end else
	begin
		iActualY := frmCMNES.Height -4{- 23};
	end;
	try
		try
		bmpCapture := TBitmap.Create;
		with bmpCapture do
			begin
				PixelFormat := pf24bit;
				//Width:=frmCMNES.Width;
				Width := iActualX - 4;
				//Height:=frmCMNES.Height;
				if NESScreen.FullScreen then
					Height := iActualY
				else
					Height := iActualY -42;
				if NESScreen.FullScreen then
					BitBlt(canvas.handle, -4, 0, iActualX, iActualY, DeskDC, frmCMNES.Left,frmCMNES.Top, SRCCOPY)
				else
					BitBlt(canvas.handle, -4, -42, iActualX, iActualY, DeskDC, frmCMNES.Left,frmCMNES.Top, SRCCOPY);
			end;
			try
				bmpCapture.SaveToFile(strScreenCapturePath + '\screen' + IntToStr(iNextFileNo) + '.bmp');
			except
				{}
			end;

		except
			{}
		end;
	finally
		ReleaseDC(GetDesktopWindow, DeskDC);
	end;

end;

procedure TfrmCMNES.mnuAboutClick(Sender: TObject);
begin
	with TfrmAbout.Create(Application) do
	begin
		try
			ShowModal;
		finally
			Free;
		end;
	end;
end;

procedure TfrmCMNES.tmrFramesPerSecondTimer(Sender: TObject);
begin
	{update frames per second and CPU speed}
	strFrameCount := IntToStr(FrameCount) + ' fps' + strCPUSpeed;
	FrameCount := 0;
end;

procedure TfrmCMNES.mnuFramesPerSecondClick(Sender: TObject);
begin
	{toggle frames per second}
	mnuFramesPerSecond.Checked := not mnuFramesPerSecond.Checked;
	tmrFramesPerSecond.Enabled := mnuFramesPerSecond.Checked;
end;

procedure TfrmCMNES.mnuSettingsClick(Sender: TObject);
begin
	frmSettings.Show;
end;

procedure TfrmCMNES.SetRegistrySettings;
var
	myReg:		TRegistry;
begin
	{not very elegant but it works for the moment}
	if frmSettings.rb640.Checked then
		strDisplayMode := '640'
	else
		if frmSettings.rb320.Checked then
			strDisplayMode := '320'
		else
			strDisplayMode := '320';

	try
		myReg := TRegistry.Create;
		{create key if need be}
		if not myReg.KeyExists(STRREGKEY) then
		begin
			myReg.CreateKey(STRREGKEY);
		end;
		myReg.OpenKey(STRREGKEY,True);
		myReg.WriteString('ROM',strROMPath);
		myReg.WriteString('ScreenCapture',strScreenCapturePath);
		myReg.WriteString('SoundCapture',strSoundCapturePath);
		myReg.WriteString('Save',strSavePath);
		myReg.WriteString('DisplayMode',strDisplayMode);
		myReg.WriteBool('Windowed',mnuWindowed.Checked);
		myReg.CloseKey;
	finally
		myReg.Free;
	end;

end;

procedure TfrmCMNES.FormClose(Sender: TObject; var Action: TCloseAction);
begin
	{save settings}
	SetRegistrySettings;
end;

procedure TfrmCMNES.GetRegistrySettings;
var
	myReg:	 TRegistry;
begin
	try
		myReg := TRegistry.Create;
		begin
			myReg.OpenKey(STRREGKEY,True);
			try
				strROMPath := myReg.ReadString('ROM');
				strScreenCapturePath := myReg.ReadString('ScreenCapture');
				strSoundCapturePath := myReg.ReadString('SoundCapture');
				strSavePath := myReg.ReadString('Save');
				strDisplayMode := myReg.ReadString('DisplayMode');
				mnuWindowed.Checked := myReg.ReadBool('Windowed');
			except
				{}
			end;
			myReg.CloseKey;
		end;
	finally
		MyReg.Free;
	end;
end;

procedure TfrmCMNES.SetFrameSkip;
begin
	iFrameskip := 0;
	with frmSettings do
	begin
		if rbfsNone.Checked then
			iSkip := 5;
		if rbfs25.Checked then
			iSkip := 1;
		if rbfs33.Checked then
			iSkip := 2;
		if rbfs50.Checked then
			iSkip := 3;
		if rbfs75.Checked then
			iSkip := 4;
	end;
end;

{ TpAPU }

function TpAPU.Initialise(): Boolean;
begin
	Result := True;
end;

function TpAPU.CH1WriteByte(Value: Byte): Boolean;
begin
	Result := True;
end;

constructor TpAPU.Create;
begin
	{}
end;

destructor TpAPU.Destroy;
begin
	{}
end;

procedure TfrmCMNES.FormShow(Sender: TObject);
begin
	if strDisplayMode = '640' then
		frmSettings.rb640.Checked := True
	else
		if strDisplayMode = '320' then
			frmSettings.rb320.Checked := True
		else
			frmSettings.rb320.Checked := True;

end;

procedure TfrmCMNES.tmrDIPollTimer(Sender: TObject);
var
	HR:	HRESULT;
begin
	if Input.pDIJoystick <> nil then
	begin
		HR := Input.pDIJoystick.Poll;

		if HR <> DI_OK then
		begin
			frmCMNES.DXError(HR);
		end;

		HR := Input.pDIJoystick.GetDeviceState(SizeOf(TDIJOYSTATE2),@Input.JoyState);

		if HR = DIERR_INPUTLOST then
		begin
			Input.pDIJoystick.Acquire;
			HR := Input.pDIJoystick.GetDeviceState(SizeOf(TDIJOYSTATE2),@Input.JoyState);
		end;

		if HR <> DI_OK then
		begin
			frmCMNES.DXError(HR);
		end;
	end;

	if (Input.pDIKeyboard <> nil) then
	begin
		HR := Input.pDIKeyboard.GetDeviceState(256, @Input.keystate);
		if HR <> DI_OK then
		begin
			frmCMNES.DXError(HR);
		end;
	end;

end;

end.
